/*
 * $Id: DoubleListC.C,v 1.1 1998/06/24 01:29:07 lang Exp $
 *
 * Copyright (c) 1992 HaL Computer Systems, Inc.  All rights reserved.
 * 
 *          HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.
 *                  1315 Dell Avenue
 *                  Campbell, CA  95008
 *
 * Author: Greg Hilton
 * Contributors: Tom Lang, Frank Bieser, and others
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * http://www.gnu.org/copyleft/gpl.html
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "HalAppC.h"
#include "DoubleListC.h"
#include "WXmString.h"

#include <Xm/ArrowB.h>
#include <Xm/List.h>
#include <Xm/Label.h>
#include <Xm/Form.h>
#include <Xm/RowColumn.h>
#include <Xm/ScrolledW.h>

/*---------------------------------------------------------------
 *  Callback routine to handle selection in the lists 
 */

void	
DoubleListC::DoList1Selection(Widget, DoubleListC *dl, XmListCallbackStruct* cbs)
{
    char*    tmpText;
    dl->list1selected.removeAll();
    dl->list1SelCount = cbs->selected_item_count;
    if (dl->list1SelCount > 0) {
      XtSetSensitive(dl->rightArrow, True);
    } else {
      XtSetSensitive(dl->rightArrow, False);
    }
    for (int i=0; i<dl->list1SelCount; i++) {
	XmStringGetLtoR(cbs->selected_items[i], XmSTRING_DEFAULT_CHARSET, &tmpText);
	dl->list1selected.append(tmpText);
    }
}

void	
DoubleListC::DoList2Selection(Widget, DoubleListC *dl, XmListCallbackStruct* cbs)
{
    char*    tmpText;
    dl->list2selected.removeAll();
    dl->list2SelCount = cbs->selected_item_count;
    if (dl->list2SelCount > 0) {
      XtSetSensitive(dl->leftArrow, True);
    } else {
      XtSetSensitive(dl->leftArrow, False);
    }
    for (int i=0; i<dl->list2SelCount; i++) {
	XmStringGetLtoR(cbs->selected_items[i], XmSTRING_DEFAULT_CHARSET, &tmpText);
	dl->list2selected.append(tmpText);
    }
}

/*---------------------------------------------------------------
 *  Callback routine to handle MOVE LEFT Arrow Selection
 */

void
DoubleListC::DoMoveLeft(Widget, DoubleListC *dl, XtPointer)
{
//
// We are going to move items from list2 to list1 and remove
// the selected names from list2
//
    StringC selectedStr;
    if(dl->list2selected.size()) {
      for(int i=0; i<dl->list2SelCount; i++) {

	selectedStr = *(dl->list2selected[i]);

	// list management
	dl->list2strs.remove(selectedStr);
	dl->list2count--;
	dl->list1strs.append(selectedStr);
	dl->list1count++;

	// visual management
	XmString str = XmStringCreateLtoR((char*)selectedStr, XmFONTLIST_DEFAULT_TAG);
	int position = dl->list1strs.orderOf(selectedStr) + 1;
	XmListAddItem(dl->list1Box, str, position);
	XmListDeleteItem(dl->list2Box, str);
	XmStringFree(str);
      }
    }
    dl->list2SelCount = 0;
    XtSetSensitive(dl->leftArrow, False);
}

/*---------------------------------------------------------------
 *  Callback routine to handle MOVE RIGHT Arrow Selection
 */

void
DoubleListC::DoMoveRight(Widget, DoubleListC *dl, XtPointer)
{
//
// We are going to move items from list1 to list2 and remove
// the selected names from list1
//
    StringC tmpStr;
    if(dl->list1selected.size()) {
      for(int i=0; i<dl->list1SelCount; i++) {

	tmpStr = *(dl->list1selected[i]);

	// list management
	dl->list1strs.remove(tmpStr);
	dl->list1count--;
	dl->list2strs.append(tmpStr);
	dl->list2count++;

	// visual management
	
	XmString str = XmStringCreateLtoR((char*)tmpStr, XmFONTLIST_DEFAULT_TAG);
	int position = dl->list2strs.orderOf(tmpStr) + 1;
	XmListAddItem(dl->list2Box, str, position);
	XmListDeleteItem(dl->list1Box, str);
	XmStringFree(str);
      }
    }

    dl->list1SelCount = 0;
    XtSetSensitive(dl->rightArrow, False);
}

/*---------------------------------------------------------------
 *  Method to set the items in the list2
 */

void
DoubleListC::SetList(Widget listBox, const StringListC& names)
{
   XmListDeleteAllItems(listBox);

   int	count = names.size();
   XmStringTable listItems = new XmString[count];
   for (int i=0; i<count; i++) {
      listItems[i] = XmStringCreateLtoR((char*)(*names[i]), 
					XmFONTLIST_DEFAULT_TAG);
   }
   XtVaSetValues(listBox, XmNitems, listItems, XmNitemCount, count, NULL);
   for (i = 0; i < count; i++) {
     XmStringFree(listItems[i]);
   }
   delete listItems;
}
/*---------------------------------------------------------------
 *  Method to set the items in the list2
 */

void
DoubleListC::SetList1(const StringListC& names)
{
    if (list1strs.size() > 0) {
      for(int i=0; i<names.size(); i++) {
	  list1strs.append(*names[i]);
      }
    } else {
      list1strs = names;
    }
    SetList(list1Box, list1strs);
}

/*---------------------------------------------------------------
 *  Method to set the items in the list2
 */

void
DoubleListC::SetList2(const StringListC& names)
{
    if (list2strs.size() > 0) {
      for(int i=0; i<names.size(); i++) {
	list2strs.append(*names[i]);
      }
    } else {
      list2strs = names;
    }
    SetList(list2Box, list2strs);
}

void
DoubleListC::MoveToList1(StringC &name)
{
//
// Begin by getting the location of the string from list2.  If it is not in
// list2, return.
   int position = list2strs.indexOf(name);
   if (position < 0) {
     return;
   }

//
// Remove from list2 and the list2Box
//
   list2strs.remove((unsigned int)position);
   if (list2selected.includes(name)) {
     list2selected.remove(name);
     list2SelCount--;
     if (list2SelCount == 0) {
       XtSetSensitive(leftArrow, False);
     }
   }
   XmListDeletePos(list2Box, position+1);

//
// Add to list1 and the list1Box
//
   list1strs.append(name);
   position = list1strs.indexOf(name);
   XmString tcs = XmStringCreateLtoR((char*)name, XmFONTLIST_DEFAULT_TAG);
   XmListAddItem(list1Box, tcs, position+1);
   XmStringFree(tcs);
}

void
DoubleListC::MoveToList2(StringC &name)
{
//
// Begin by getting the location of the string from list1.  If it is not in
// list1, return.
   int position = list1strs.indexOf(name);
   if (position < 0) {
     return;
   }

//
// Remove from list1 and the list1Box
//
   list1strs.remove((unsigned int)position);
   if (list1selected.includes(name)) {
     list1selected.remove(name);
     list1SelCount--;
     if (list1SelCount == 0) {
       XtSetSensitive(rightArrow, False);
     }
   }
   XmListDeletePos(list1Box, position+1);

//
// Add to list2 and the list1Box
//
   list2strs.append(name);
   position = list2strs.indexOf(name);
   XmString tcs = XmStringCreateLtoR((char*)name, XmFONTLIST_DEFAULT_TAG);
   XmListAddItem(list2Box, tcs, position+1);
   XmStringFree(tcs);
}

void
DoubleListC::DeleteFromLists(StringC &name)
{
  int position = list1strs.indexOf(name);
  if (position >= 0) {
    list1strs.remove((unsigned int)position);
    XmListDeletePos(list1Box, position+1);
    if (list1selected.includes(name)) {
      list1selected.remove(name);
      list1SelCount--;
      if (list1SelCount == 0) {
	XtSetSensitive(rightArrow, False);
      }
    }
  } else {
    position = list2strs.indexOf(name);
    if (position >= 0) {
      list2strs.remove((unsigned int)position);
      XmListDeletePos(list2Box, position+1);
      if (list2selected.includes(name)) {
	list2selected.remove(name);
	list2SelCount--;
	if (list2SelCount == 0) {
	  XtSetSensitive(leftArrow, False);
        }
      }
    }
  }
}

void
DoubleListC::AddToList1(StringC &name)
{
  if (list1strs.includes(name)) {
    return;
  }

  list1strs.append(name);
  int position = list1strs.indexOf(name);
  XmString tcs = XmStringCreateLtoR((char*)name, XmFONTLIST_DEFAULT_TAG);
  XmListAddItem(list1Box, tcs, position+1);
  XmStringFree(tcs);
}

void
DoubleListC::AddToList2(StringC &name)
{
  if (list2strs.includes(name)) {
    return;
  }

  list2strs.append(name);
  int position = list2strs.indexOf(name);
  XmString tcs = XmStringCreateLtoR((char*)name, XmFONTLIST_DEFAULT_TAG);
  XmListAddItem(list2Box, tcs, position+1);
  XmStringFree(tcs);
}

/*---------------------------------------------------------------
 *  Method to get the items from the list1
 */

StringListC&
DoubleListC::GetList(Widget listBox, StringListC& strItems)
{
// pass in a StringList and the item list is appended
//
   int	itemCount;
   XmString	*items;
   XtVaGetValues(listBox,
		 XmNitemCount, &itemCount, 
		 XmNitems,     &items, 
		 NULL);
   for(int i=0; i < itemCount; i++) {
       char	*str;
       XmStringGetLtoR(items[i], XmFONTLIST_DEFAULT_TAG, &str);
       strItems.append(str);
       XtFree(str);
   }
   return(strItems);
}

StringListC&
DoubleListC::GetList1(StringListC& strItems)
{
    return( GetList(list1Box, strItems));
}

/*---------------------------------------------------------------
 *  Method to get the items from the list2
 */

StringListC&
DoubleListC::GetList2(StringListC& strItems)
{
    return( GetList(list2Box, strItems));
}

/*---------------------------------------------------------------
 *  Method to remove all items from both of the lists
 */

void
DoubleListC::Clear()
{
   XmListDeleteAllItems(list1Box);
   XmListDeleteAllItems(list2Box);
}


/*---------------------------------------------------------------
 *  Constructor
 */

void
DoubleListC::Build(Widget parent, const StringC& name)
{
   WArgList	args;

   list1count = 0;
   list2count = 0;

/*---------------------------------------------------------------
 *  Create the main form
 */

   StringC tempStr = name;
   args.Reset();
   args.TopAttachment(XmATTACH_FORM);
   args.LeftAttachment(XmATTACH_FORM);
   args.RightAttachment(XmATTACH_FORM);
   args.BottomAttachment(XmATTACH_FORM);
   mainForm = XmCreateForm(parent, (char*)tempStr, ARGS);

// DJL - This has been changed to keep the arrows the same size and attached
// to the middle of the form.
//
// This is the hook.  Start by creating the list forms and attach them to 
// position 50.  Add the arrow form, placing it in between the two.  Then,
// determine how big the arrows are and set offsets for the list forms so that
// they leave room for the arrow form.
//

   args.Reset();
   args.TopAttachment(XmATTACH_FORM);
   args.LeftAttachment(XmATTACH_FORM);
   args.RightAttachment(XmATTACH_POSITION, 50);
   args.BottomAttachment(XmATTACH_FORM);
   Widget list1Form   = XmCreateForm(mainForm, "list1Form", ARGS);

   args.Reset();
   args.TopAttachment(XmATTACH_FORM);
   args.LeftAttachment(XmATTACH_POSITION, 50);
   args.RightAttachment(XmATTACH_FORM);
   args.BottomAttachment(XmATTACH_FORM);
   Widget list2Form   = XmCreateForm(mainForm, "list2Form", ARGS);

   args.Reset();
   args.TopAttachment(XmATTACH_FORM);
   args.LeftAttachment(XmATTACH_WIDGET, list1Form);
   args.RightAttachment(XmATTACH_WIDGET, list2Form);
   args.BottomAttachment(XmATTACH_FORM);
   Widget arrowForm   = XmCreateForm(mainForm, "arrowForm", ARGS);

/*---------------------------------------------------------------
 *  Create arrow buttons
 */

   args.Reset();
   args.Packing(XmPACK_COLUMN);
   args.Orientation(XmVERTICAL);
   args.IsAligned(True);
   args.EntryAlignment(XmALIGNMENT_CENTER);
   args.LeftAttachment(XmATTACH_FORM);
   args.TopAttachment(XmATTACH_FORM);
   args.TopOffset(20);
   args.RightAttachment(XmATTACH_FORM);
   args.BottomAttachment(XmATTACH_FORM);
   Widget arrowBox = XmCreateRowColumn(arrowForm, "arrowBox", ARGS);
   XtManageChild(arrowBox);

   args.Reset();
   args.ArrowDirection(XmARROW_RIGHT);
   rightArrow = XmCreateArrowButton(arrowBox, "rightArrow", ARGS);
   XtManageChild(rightArrow);
   XtAddCallback(rightArrow,XmNactivateCallback,
		 (XtCallbackProc)DoubleListC::DoMoveRight, (caddr_t)this);
   XtSetSensitive(rightArrow, False);

   args.Reset();
   args.ArrowDirection(XmARROW_LEFT);
   leftArrow = XmCreateArrowButton(arrowBox, "leftArrow", ARGS);
   XtManageChild(leftArrow);
   XtAddCallback(leftArrow,XmNactivateCallback,
		 (XtCallbackProc)DoubleListC::DoMoveLeft, (caddr_t)this);
   XtSetSensitive(leftArrow, False);

   Dimension width;
   XtVaGetValues(leftArrow, XmNwidth, &width, NULL);
   width >>= 1;
   int offst = (int)width + 4;
   XtVaSetValues(list2Form, XmNleftOffset, offst, NULL);
   XtVaSetValues(list1Form, XmNrightOffset, offst, NULL);
  
   args.Reset();
   args.LeftOffset((int)width+4);
   args.RightOffset(-((int)width + 4));

   XtManageChild(mainForm);
   XtManageChild(list1Form);
   XtManageChild(arrowForm);
   XtManageChild(list2Form);

/*---------------------------------------------------------------
 *  Create first scrolled list
 */

   list1Title = XmCreateLabel(list1Form, "list1Title", 0,0);
   XtManageChild(list1Title);

   args.Reset();
   args.SelectionPolicy(XmEXTENDED_SELECT);
   args.VisibleItemCount(5);
#if !defined(SUN_OS) & !defined(SOLARIS)
   args.ScrollBarDisplayPolicy(XmSTATIC);
   args.ListSizePolicy(XmCONSTANT);
#endif
   args.LeftAttachment(XmATTACH_FORM);
   args.RightAttachment(XmATTACH_FORM);
   args.TopAttachment(XmATTACH_WIDGET, list1Title);
   args.BottomAttachment(XmATTACH_FORM);
   list1Box = XmCreateScrolledList(list1Form, "list1Box", ARGS);
   XtManageChild(list1Box);
   XtAddCallback(list1Box, XmNextendedSelectionCallback, 
		 (XtCallbackProc)DoubleListC::DoList1Selection, (caddr_t)this);
   XtAddCallback(list1Box, XmNdefaultActionCallback, 
		 (XtCallbackProc)DoubleListC::DoList1Selection, (caddr_t)this);


/*---------------------------------------------------------------
 *  Create second scrolled list
 */

   list2Title = XmCreateLabel(list2Form, "list2Title", 0,0);
   XtManageChild(list2Title);

   args.Reset();
   args.SelectionPolicy(XmEXTENDED_SELECT);
   args.VisibleItemCount(5);
#if !defined(SUN_OS) & !defined(SOLARIS)
   args.ScrollBarDisplayPolicy(XmSTATIC);
   args.ListSizePolicy(XmCONSTANT);
#endif
   args.LeftAttachment(XmATTACH_FORM);
   args.RightAttachment(XmATTACH_FORM );
   args.TopAttachment(XmATTACH_WIDGET, list2Title);
   args.BottomAttachment(XmATTACH_FORM);
   list2Box = XmCreateScrolledList(list2Form, "list2Box", ARGS);
   XtManageChild(list2Box);
   XtAddCallback(list2Box, XmNextendedSelectionCallback, 
		 (XtCallbackProc)DoubleListC::DoList2Selection, (caddr_t)this);
   XtAddCallback(list2Box, XmNdefaultActionCallback, 
		 (XtCallbackProc)DoubleListC::DoList2Selection, (caddr_t)this);

} // End DoubleListC::DoubleListC

/*---------------------------------------------------------------
 *  Destructor
 */

DoubleListC::~DoubleListC()
{
   if ( halApp->xRunning )
      XtDestroyWidget(mainForm);
}

