//
// GraphListC.h
//

#ifndef _GraphListC_h_
#define _GraphListC_h_

#include "GraphC.h"

/*
 * PListC.clas -- class definition file for class PListC
 *
 *   This type of list stores only pointers to its members
 *   OListC stores the members themselves
 */

class GraphListC {

protected:

   GraphC			**_list;
   unsigned		_count;
   unsigned		_space;
   char			allow_duplicates;
   char			sorted;
   char         	adjustGrowth;
   char         	autoShrink;
   void			grow(unsigned size=0);
   void			shrink();

public:

   unsigned		GROWTH_AMOUNT;
   const int		NULL_INDEX;
   const unsigned	MAX_GROWTH;

   GraphListC();
   GraphListC(unsigned ga);
   GraphListC(const GraphListC&);
   ~GraphListC();

/* Append another list to this one */
   GraphListC& operator+=(const GraphListC&);

/* Copy another list to this one */
   GraphListC& operator=(const GraphListC&);

/* Return the position of the specified object.  Object must be in list */

   int		indexOf(const GraphC&) const;
   int		indexOf(const GraphC*) const;

/* Return the sorted position of the specified object.  Object does not */
/*    have to be in the list */

   unsigned		orderOf(const GraphC&) const;
   unsigned		orderOf(const GraphC*) const;

/* Return non-zero if the specified object is in the list */

   int		includes(const GraphC&) const;
   int		includes(const GraphC*) const;

/* Add the specified object to the list in the requested position */
/*    The position is ignored if this list is sorted */

   GraphC		*insert (const GraphC*, unsigned);
   GraphC		*replace(const GraphC*, unsigned);
   GraphC		*append (const GraphC*);
   GraphC		*prepend(const GraphC*);
   GraphC		*add    (const GraphC*);

/* Remove the specified object or index from the list */

   void		remove(unsigned);
   void		remove(const GraphC&);
   void		remove(const GraphC*);
   void		removeNulls();
   void		removeAll();
   void		removeLast();

/* Return a pointer to the previous object in the list */

   GraphC		*before(const GraphC*) const;

/* Functions to support sorting */

   static int	compare(const void*, const void*);
   void		sort();
   void		sort(int (*)(const void*, const void*));
   void		SetSorted(char);

   inline unsigned	size()		const { return _count; }
   inline unsigned	capacity()	const { return _space; }
   inline GraphC		**start()	const { return _list; }
   void         	AllowDuplicates(char);
   void			AutoShrink(char);
   void			SetCapacity(unsigned);

   GraphC			*operator[](unsigned) const;
   GraphC			*First() const;
   GraphC			*Last() const;

   ostream&		printOn(ostream& strm=cout) const;
};

/*-------------------------------------------------------------*/
/* PRINTING */

inline ostream&
operator<<(ostream& strm, const GraphListC& c)
{
   return c.printOn(strm);
}

#endif // _GraphListC_h_
 * 
 *          HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.
 *                  1315 Dell Avenue
 *                  Campbell, CA  95008
 *
 * Author: Greg Hilton
 * Contributors: Tom Lang, Frank Bieser, and others
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * http://www.gnu.org/copyleft/gpl.html
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
