//
// LayoutListC.h
//

#ifndef _LayoutListC_h_
#define _LayoutListC_h_

#include "LayoutC.h"

/*
 * PListC.clas -- class definition file for class PListC
 *
 *   This type of list stores only pointers to its members
 *   OListC stores the members themselves
 */

class LayoutListC {

protected:

   LayoutC			**_list;
   unsigned		_count;
   unsigned		_space;
   char			allow_duplicates;
   char			sorted;
   char         	adjustGrowth;
   char         	autoShrink;
   void			grow(unsigned size=0);
   void			shrink();

public:

   unsigned		GROWTH_AMOUNT;
   const int		NULL_INDEX;
   const unsigned	MAX_GROWTH;

   LayoutListC();
   LayoutListC(unsigned ga);
   LayoutListC(const LayoutListC&);
   ~LayoutListC();

/* Append another list to this one */
   LayoutListC& operator+=(const LayoutListC&);

/* Copy another list to this one */
   LayoutListC& operator=(const LayoutListC&);

/* Return the position of the specified object.  Object must be in list */

   int		indexOf(const LayoutC&) const;
   int		indexOf(const LayoutC*) const;

/* Return the sorted position of the specified object.  Object does not */
/*    have to be in the list */

   unsigned		orderOf(const LayoutC&) const;
   unsigned		orderOf(const LayoutC*) const;

/* Return non-zero if the specified object is in the list */

   int		includes(const LayoutC&) const;
   int		includes(const LayoutC*) const;

/* Add the specified object to the list in the requested position */
/*    The position is ignored if this list is sorted */

   LayoutC		*insert (const LayoutC*, unsigned);
   LayoutC		*replace(const LayoutC*, unsigned);
   LayoutC		*append (const LayoutC*);
   LayoutC		*prepend(const LayoutC*);
   LayoutC		*add    (const LayoutC*);

/* Remove the specified object or index from the list */

   void		remove(unsigned);
   void		remove(const LayoutC&);
   void		remove(const LayoutC*);
   void		removeNulls();
   void		removeAll();
   void		removeLast();

/* Return a pointer to the previous object in the list */

   LayoutC		*before(const LayoutC*) const;

/* Functions to support sorting */

   static int	compare(const void*, const void*);
   void		sort();
   void		sort(int (*)(const void*, const void*));
   void		SetSorted(char);

   inline unsigned	size()		const { return _count; }
   inline unsigned	capacity()	const { return _space; }
   inline LayoutC		**start()	const { return _list; }
   void         	AllowDuplicates(char);
   void			AutoShrink(char);
   void			SetCapacity(unsigned);

   LayoutC			*operator[](unsigned) const;
   LayoutC			*First() const;
   LayoutC			*Last() const;

   ostream&		printOn(ostream& strm=cout) const;
};

/*-------------------------------------------------------------*/
/* PRINTING */

inline ostream&
operator<<(ostream& strm, const LayoutListC& c)
{
   return c.printOn(strm);
}

#endif // _LayoutListC_h_
 * 
 *          HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.
 *                  1315 Dell Avenue
 *                  Campbell, CA  95008
 *
 * Author: Greg Hilton
 * Contributors: Tom Lang, Frank Bieser, and others
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * http://www.gnu.org/copyleft/gpl.html
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
