//
// MimeRichTextListC.h
//

#ifndef _MimeRichTextListC_h_
#define _MimeRichTextListC_h_

#include "MimeRichTextC.h"

/*
 * PListC.clas -- class definition file for class PListC
 *
 *   This type of list stores only pointers to its members
 *   OListC stores the members themselves
 */

class MimeRichTextListC {

protected:

   MimeRichTextC			**_list;
   unsigned		_count;
   unsigned		_space;
   char			allow_duplicates;
   char			sorted;
   char         	adjustGrowth;
   char         	autoShrink;
   void			grow(unsigned size=0);
   void			shrink();

public:

   unsigned		GROWTH_AMOUNT;
   const int		NULL_INDEX;
   const unsigned	MAX_GROWTH;

   MimeRichTextListC();
   MimeRichTextListC(unsigned ga);
   MimeRichTextListC(const MimeRichTextListC&);
   ~MimeRichTextListC();

/* Append another list to this one */
   MimeRichTextListC& operator+=(const MimeRichTextListC&);

/* Copy another list to this one */
   MimeRichTextListC& operator=(const MimeRichTextListC&);

/* Return the position of the specified object.  Object must be in list */

   int		indexOf(const MimeRichTextC&) const;
   int		indexOf(const MimeRichTextC*) const;

/* Return the sorted position of the specified object.  Object does not */
/*    have to be in the list */

   unsigned		orderOf(const MimeRichTextC&) const;
   unsigned		orderOf(const MimeRichTextC*) const;

/* Return non-zero if the specified object is in the list */

   int		includes(const MimeRichTextC&) const;
   int		includes(const MimeRichTextC*) const;

/* Add the specified object to the list in the requested position */
/*    The position is ignored if this list is sorted */

   MimeRichTextC		*insert (const MimeRichTextC*, unsigned);
   MimeRichTextC		*replace(const MimeRichTextC*, unsigned);
   MimeRichTextC		*append (const MimeRichTextC*);
   MimeRichTextC		*prepend(const MimeRichTextC*);
   MimeRichTextC		*add    (const MimeRichTextC*);

/* Remove the specified object or index from the list */

   void		remove(unsigned);
   void		remove(const MimeRichTextC&);
   void		remove(const MimeRichTextC*);
   void		removeNulls();
   void		removeAll();
   void		removeLast();

/* Return a pointer to the previous object in the list */

   MimeRichTextC		*before(const MimeRichTextC*) const;

/* Functions to support sorting */

   static int	compare(const void*, const void*);
   void		sort();
   void		sort(int (*)(const void*, const void*));
   void		SetSorted(char);

   inline unsigned	size()		const { return _count; }
   inline unsigned	capacity()	const { return _space; }
   inline MimeRichTextC		**start()	const { return _list; }
   void         	AllowDuplicates(char);
   void			AutoShrink(char);
   void			SetCapacity(unsigned);

   MimeRichTextC			*operator[](unsigned) const;
   MimeRichTextC			*First() const;
   MimeRichTextC			*Last() const;

   ostream&		printOn(ostream& strm=cout) const;
};

/*-------------------------------------------------------------*/
/* PRINTING */

inline ostream&
operator<<(ostream& strm, const MimeRichTextListC& c)
{
   return c.printOn(strm);
}

#endif // _MimeRichTextListC_h_
