/*
 *  $Id: TreeDataC.C,v 1.1 1998/06/24 01:29:07 lang Exp $
 *  
 *  Copyright (c) 1994 HAL Computer Systems International, Ltd.
 * 
 *          HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.
 *                  1315 Dell Avenue
 *                  Campbell, CA  95008
 *
 * Author: Greg Hilton
 * Contributors: Tom Lang, Frank Bieser, and others
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * http://www.gnu.org/copyleft/gpl.html
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#include "TreeDataC.h"
#include "VItemC.h"

TreeDataC::TreeDataC()
{
   item    = NULL;
   view    = NULL;
   depth   = 0;
   visible = True;
}

TreeDataC::TreeDataC(const VItemC& i, TreeViewC* tv)
{
   item    = (VItemC *)&i;
   view    = tv;
   depth   = 0;
   visible = True;
}

void
TreeDataC::operator=(const TreeDataC& d)
{
   item    = d.item;
   view    = d.view;
   bounds  = d.bounds;
   domain  = d.domain;
   depth   = d.depth;
   visible = d.visible;
}

Boolean
TreeDataC::Visible() const
{
   if ( !visible ) return False;

//
// Even if this one is visible, we must check ancestors
//
   VItemC	*parent = item->parent;
   while ( parent ) {
      TreeDataC	*data = (TreeDataC*)parent->ViewData();
      if ( !data->visible ) return False;
      parent = parent->parent;
   }

//
// All ancestors visible
//
   return True;
}

VItemC*
TreeDataC::PrevVisible() const
{
   VItemC	*prev = item->prev;
   while ( prev ) {
      TreeDataC	*data = (TreeDataC*)prev->ViewData();
      if ( data && data->Visible() ) return prev;
      prev = prev->prev;
   }

   return NULL;
}

VItemC*
TreeDataC::NextVisible() const
{
   VItemC	*next = item->next;
   while ( next ) {
      TreeDataC	*data = (TreeDataC*)next->ViewData();
      if ( data && data->Visible() ) return next;
      next = next->next;
   }

   return NULL;
}
