/*
 *  $Id: TreeViewC.C,v 1.1 1998/06/24 01:29:07 lang Exp $
 *  
 *  Copyright (c) 1994 HAL Computer Systems International, Ltd.
 * 
 *          HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.
 *                  1315 Dell Avenue
 *                  Campbell, CA  95008
 *
 * Author: Greg Hilton
 * Contributors: Tom Lang, Frank Bieser, and others
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * http://www.gnu.org/copyleft/gpl.html
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#include "TreeViewC.h"
#include "TreeDataC.h"
#include "HalAppC.h"
#include "VBoxC.h"
#include "rsrc.h"
#include "WArgList.h"
#include "CharC.h"

#include <Xm/ScrollBar.h>
#include <Xm/AtomMgr.h>
#include <Xm/DragDrop.h>

#include <X11/keysym.h>

#include <values.h>
#include <math.h>

#define WHITE	WhitePixel(halApp->display, DefaultScreen(halApp->display))

extern int	debuglev;

/*-----------------------------------------------------------------------
 *  Return widget name for button in popup menu used to display this view
 */

char *
TreeViewC::ButtonName() const
{
   return "treeViewPB";
}

/*-----------------------------------------------------------------------
 *  Method to check data pointer
 */

inline Boolean
TreeViewC::DataValid(const TreeDataC *data)
{
   return ( data && (data->view == this) );
}

/*-----------------------------------------------------------------------
 *  Method to check visibility
 */

inline Boolean
TreeViewC::DataVisible(const TreeDataC *data)
{
   return ( DataValid(data) && data->Visible() );
}

/*-----------------------------------------------------------------------
 *  Constructor for TreeViewC
 */

TreeViewC::TreeViewC(VBoxC *vb) : ViewC(vb)
{
   needWd        =
   needHt        = 0;
   focusHere	 = False;
   hlItem	 = NULL;
   flashItem	 = NULL;
   flashTimer	 = (XtIntervalId)NULL;
   dropItem	 = NULL;
   dragIcon      = NULL;
   pickGC        = NULL;
   scrollTimer   = (XtIntervalId)NULL;
   pressItem     = NULL;
   clickCount    = 1;
   lastClickTime = 0;
   needPlaces    = True;

   Pixel	fg;
   XtVaGetValues(viewBox->ViewDA(), XmNforeground, &fg, NULL);

//
// Get resources
//
   StringC	tmp;
   Widget	da = viewBox->ViewDA();
   char		*cl = "TreeViewC";
   bgColor      = get_color  (cl, da, "background",		WHITE);
   lineColor    = get_color  (cl, da, "lineColor",		fg);
   etchColor    = get_color  (cl, da, "topShadowColor",		WHITE);
   hlColor      = get_color  (cl, da, "highlightColor",		fg);
   dropColor    = get_color  (cl, da, "validDropColor",		hlColor);
   hlThick      = get_int    (cl, da, XmNhighlightThickness,	0);
   marginWd	= get_int    (cl, da, XmNmarginWidth,		5);
   marginHt	= get_int    (cl, da, XmNmarginHeight,		5);
   genSpacing	= get_int    (cl, da, "generationSpacing",	20);
   sibSpacing	= get_int    (cl, da, "siblingSpacing",		10);
   lineWidth	= get_int    (cl, da, "lineWidth",		0);
   compress	= get_boolean(cl, da, "compress",		True);

   rootGravity	= get_gravity(cl, da, "rootGravity",		"West");
   if ( rootGravity != NorthGravity ) rootGravity = WestGravity;

   tmp = get_string(cl, da, "treeStyle", "Angle");
   if      ( tmp.StartsWith("outline", IGNORE_CASE) ) treeStyle = TREE_OUTLINE;
   else if ( tmp.StartsWith("square",  IGNORE_CASE) ) treeStyle = TREE_SQUARED;
   else						      treeStyle = TREE_ANGLED;

   tmp = get_string(cl, da, "lineStyle", "Solid");
   if      ( tmp.Equals("etched_in",  IGNORE_CASE) )
      lineStyle = XmSHADOW_ETCHED_IN;
   else if ( tmp.Equals("dashed_in",  IGNORE_CASE) )
      lineStyle = XmSHADOW_ETCHED_IN_DASH;
   else if ( tmp.Equals("etched_out", IGNORE_CASE) )
      lineStyle = XmSHADOW_ETCHED_OUT;
   else if ( tmp.Equals("dashed_out", IGNORE_CASE) )
      lineStyle = XmSHADOW_ETCHED_OUT_DASH;
   else if ( tmp.Equals("dashed",     IGNORE_CASE) )
      lineStyle = XmSINGLE_DASHED_LINE;
   else
      lineStyle = XmSINGLE_LINE;

//
// Get keyboard focus policy
//
   unsigned char	focusPolicy;
   Widget		shell = XtParent(vb->ViewDA());
   while ( !XtIsShell(shell) ) shell = XtParent(shell);
   XtVaGetValues(shell, XmNkeyboardFocusPolicy, &focusPolicy, NULL);

   if ( focusPolicy == XmEXPLICIT ) {

//
// Add event handler for keyboard focus change
//
      XtAddEventHandler(vb->ViewDA(), FocusChangeMask, False,
			(XtEventHandler)HandleFocusChange, (XtPointer)this);

   } else { // XmPOINTER

      XtAddEventHandler(vb->ViewDA(), EnterWindowMask|LeaveWindowMask,
			False, (XtEventHandler)HandleFocusChange,
			(XtPointer)this);
      XtAddEventHandler(vb->ViewDA(),
			PointerMotionMask|PointerMotionHintMask, False,
			(XtEventHandler)HandlePointerMotion,
			(XtPointer)this);
   }

} // End Constructor

/*-----------------------------------------------------------------------
 *  Destructor for TreeViewC
 */

TreeViewC::~TreeViewC()
{
//
// Delete data
//
   u_int	count = dataDict.size();
   for (int i=0; i<count; i++) {
      delete dataDict[i]->val;
      delete dataDict[i];
   }

   if ( halApp->xRunning ) {
      if ( pickGC ) XtReleaseGC(halApp->appShell, pickGC);
   }

} // End Destructor

/*-----------------------------------------------------------------------
 *  Display this view
 */

void
TreeViewC::Show()
{
   shown = True;

//
// Add any items that aren't yet added
//
   VItemListC&	items = viewBox->Items();
   unsigned	count = items.size();
   for (int i=0; i<count; i++) {
      VItemC	*item = items[i];

//
// Look first at the view data, then in the dictionary
//
      TreeDataC	*data = (TreeDataC*)dataDict.definitionOf(item);
      if ( data )
	 item->SetViewData(data);
      else
	 _AddItem(*item);

   } // End for each item in view

   needPlaces = True;

//
// Set the GC values
//
   XtGCMask     valMask = GCClipMask | GCFunction | GCGraphicsExposures |
   			  GCLineStyle | GCLineWidth | GCPlaneMask |
			  GCForeground;
   XGCValues vals;
   vals.clip_mask		= None;
   vals.function		= GXcopy;
   vals.graphics_exposures	= TRUE;
   vals.line_style		= LineSolid;
   vals.line_width		= lineWidth;
   vals.plane_mask		= AllPlanes;
   vals.foreground		= lineColor;

   XChangeGC(halApp->display, viewBox->ViewGC(), valMask, &vals);

//
// Create a GC for the rubberband rectangle if necessary
//
   if ( !pickGC ) {
      Pixel	fg, bg;
      XtVaGetValues(viewBox->ViewDA(), XmNforeground, &fg,
      				       XmNbackground, &bg, NULL);

      XGCValues	pickVals;
      pickVals.foreground = fg ^ bg;
      pickVals.function   = GXxor;
      pickGC = XtGetGC(viewBox->ViewDA(), GCForeground|GCFunction, &pickVals);
   }

   Redraw();

} // End Show

/*-----------------------------------------------------------------------
 *  Remove this view
 */

void
TreeViewC::Hide()
{
//
// Clear view data
//
   VItemListC&	items = viewBox->Items();
   unsigned	count = items.size();
   for (int i=0; i<count; i++) {
      //cout <<"Clearing data for item " <<hex <<(int)items[i] <<dec NL;
      items[i]->SetViewData(NULL);
   }
   
   shown = False;

} // End Hide

/*-----------------------------------------------------------------------
 *  Public method to add a new item to this view box
 */

void
TreeViewC::AddItem(VItemC& item)
{
   if ( !viewBox->Realized() ) return;

//
// Add the item
//
   _AddItem(item);

//
// Redraw if necessary
//
   if ( !viewBox->Deferred() ) Redraw();
   else			       viewBox->Changed(True);

} // End AddItem

/*-----------------------------------------------------------------------
 *  Add new items to this view
 */

void
TreeViewC::AddItems(const VItemListC& list)
{
   if ( !viewBox->Realized() ) return;

//
// Add items
//
   unsigned	count = list.size();
   for (int i=0; i<count; i++) _AddItem(*list[i]);

//
// Redraw if necessary
//
   if ( !viewBox->Deferred() ) Redraw();
   else			       viewBox->Changed(True);

} // End AddItems

/*-----------------------------------------------------------------------
 *  Private method to add a new item to this view box
 */

void
TreeViewC::_AddItem(VItemC& item)
{
//
// Create a new icon data structure and add it to the dictionary
//
   TreeDataC	*data = new TreeDataC(item, this);
   //cout <<"Created data " <<hex <<(int)data <<" for item "
   //	  <<(int)&item SP (int)data->item <<dec NL;
   dataDict.add(&item, data);
   item.SetViewData(data);

   needPlaces = True;

} // End _AddItem

/*-----------------------------------------------------------------------
 *  Calculate positions
 */

void
TreeViewC::PlaceItems()
{
//
// Make a pass and get sizes
//
   VItemC	*item;
   TreeDataC	*data;
   VItemListC&	visItems = viewBox->VisItems();
   u_int	count    = visItems.size();
   for (int i=0; i<count; i++) {

      item = visItems[i];
      data = (TreeDataC *)item->ViewData();
      if ( !DataVisible(data) ) continue;

      if ( data->bounds.wd == 0 ) {

	 int	wd, ht;
	 item->GetPrefSize(&wd, &ht);

//
// Add highlight thickness
//
	 wd += hlThick*2;
	 ht += hlThick*2;

	 data->bounds.SetSize(wd, ht);
      }
   }

//
// Place children of top-level items
//
   int		x = marginWd;
   int		y = marginHt;
   RectC	daBounds(0, 0, 0, 0);
   for (i=0; i<count; i++) {

      item = visItems[i];
      data = (TreeDataC *)item->ViewData();
      if ( item->parent || !DataVisible(data) ) continue;

//
// Place item in upper left corner of domain for now
//
      data->bounds.SetPos(x, y);

//
//  The domain is the bounding box around the entire sub-tree
//
      data->domain = data->bounds;
      PlaceItem(item);

//
// For all except the outline layout, center this object with respect to
//    its children.
//
      if ( treeStyle != TREE_OUTLINE ) {

	 CenterItem(item);

//
// Now that we know our size, we may be able to squeeze more closely to the
//    previous tree.
//
	 CheckForBetterFit(data);
      }

//
// Add the domain of this item to the total size of the drawing area
//
      daBounds = daBounds & data->domain;

//
// Move to the start of the next top-level item
//
      if ( treeStyle == TREE_OUTLINE || rootGravity == WestGravity )
	 y = daBounds.ymax + sibSpacing;
      else
	 x = daBounds.xmax + sibSpacing;

   } // End for each visible item

   needWd = daBounds.wd + marginWd;
   needHt = daBounds.ht + marginHt;

   needPlaces = False;

} // End PlaceItems

/*-----------------------------------------------------------------------
 *  Lay out an item and its children
 */

void
TreeViewC::PlaceItem(VItemC *item)
{
   TreeDataC	*data = (TreeDataC *)item->ViewData();
   if ( !item->child || !DataVisible(data) ) return;

   int	x, y;
   if ( treeStyle == TREE_OUTLINE ) {
      x = data->bounds.xmax + genSpacing;
      y = data->bounds.ymax + sibSpacing;
   }
   else if ( rootGravity == NorthGravity ) {
      x = data->bounds.xmin;
      y = data->bounds.ymax + genSpacing;
   }
   else {
      x = data->bounds.xmax + genSpacing;
      y = data->bounds.ymin;
   }

   VItemC	*child   = item->child;
   TreeDataC	*childData;
   RectC	childBounds(x, y, 0, 0);	// Sum over all children
   while ( child ) {

//
// Process child only if visible
//
      childData = (TreeDataC *)child->ViewData();
      if ( DataVisible(childData) ) {

//
// Place child in upper left corner of domain for now
//
	 childData->bounds.SetPos(x, y);

//
//  The domain is the bounding box around the entire sub-tree
//
	 childData->domain = childData->bounds;
	 PlaceItem(child);

//
// For all except the outline layout, center this object with respect to
//    its children.
//
	 if ( treeStyle != TREE_OUTLINE ) {
	    
	    CenterItem(child);

//
// Now that we know our size, we may be able to squeeze more closely to the
//    previous tree.
//
	    CheckForBetterFit(childData);
	 }

//
// Add the domain of the child to the domain of the parent
//
	 childBounds = childBounds & childData->domain;

//
// Move to the start of the next child
//
	 if ( treeStyle == TREE_OUTLINE || rootGravity == WestGravity )
	    y = childBounds.ymax + sibSpacing;
	 else
	    x = childBounds.xmax + sibSpacing;

      } // End if data valid

      child = child->next;

   } // End for each item

//
// Merge children with parent
//
   if ( childBounds.wd > 0 )
      data->domain = data->bounds & childBounds;

} // End PlaceItem

/*-----------------------------------------------------------------------
 * See if this tree can be placed closer to the previous tree.
 *    Each tree has a sub-domain for each generation.   The 0th generation
 *    sub-domain is the bounding box of the root item.  The 1st generation
 *    sub-domain is the domain of the root item and its children.  The 2nd
 *    generation sub-domain is the domain of the root item, its children
 *    and its grand-children, etc.  The normal domain is just the Nth
 *    generation domain where N is the total number of generations.
 * We may be able to squeeze in next to one of the sub-domains of the
 *    previous tree.
 * We must check all siblings since the previous one may also be squeezed
 */

void
TreeViewC::CheckForBetterFit(TreeDataC *data)
{
   if ( !compress ) return;

   VItemC	*prev = data->PrevVisible();
   RectC	newDomain;
   RectC	bestDomain;

   while ( prev ) {

      GetClosestFit(data, prev, &newDomain);

//
// If there is no change, we're done
//
      if ( data->domain.xmin == newDomain.xmin &&
	   data->domain.ymin == newDomain.ymin )
	 return;

//
// If there is a change, save this domain
//
      if ( bestDomain.wd == 0 )
	 bestDomain = newDomain;
      else if ( treeStyle == TREE_OUTLINE || rootGravity == WestGravity ) {
	 if ( newDomain.ymin > bestDomain.ymin ) bestDomain = newDomain;
      }
      else {
	 if ( newDomain.xmin > bestDomain.xmin ) bestDomain = newDomain;
      }

      TreeDataC	*prevData = (TreeDataC*)prev->ViewData();
      prev = prevData->PrevVisible();
   }

//
// If we get here, there is a change
//
   if ( bestDomain.wd > 0 ) {
      if ( treeStyle == TREE_OUTLINE || rootGravity == WestGravity )
	 SetDomainPos(data, data->domain.xmin, bestDomain.ymin);
      else
	 SetDomainPos(data, bestDomain.xmin, data->domain.ymin);
   }

} // End CheckForBetterFit

/*-----------------------------------------------------------------------
 * Find the domain that represents the closest we can fit to the specified
 *   sibling.
 */

void
TreeViewC::GetClosestFit(TreeDataC *data, VItemC *prev, RectC *newDomain)
{
   *newDomain = data->domain;

//
// First see if we can fix next to one of the previous item's generations
//
   int		genCount = GenerationCount(prev);
   RectC	subDomain;
   for (int g=0; g<=genCount; g++) {

      GetDomain(prev, g, &subDomain);

//
// If we're stacking vertically, see if our tree will fit horizontally
//   next to this sub-domain
//
      if ( treeStyle == TREE_OUTLINE || rootGravity == WestGravity ) {
	 if ( data->domain.wd <= subDomain.wd ) {
	    *newDomain = data->domain;
	    newDomain->SetPos(data->domain.xmin, subDomain.ymax + sibSpacing);
	    return;
	 }
      }

//
// If we're stacking horizontally, see if our tree will fit vertically
//   next to this sub-domain
//
      else if ( data->domain.ht <= subDomain.ht ) {
	 *newDomain = data->domain;
	 newDomain->SetPos(subDomain.xmax + sibSpacing, data->domain.ymin);
	 return;
      }

   } // End for each generation

//
// Next see if we can fit one of our generations next to the previous item
//
   TreeDataC	*prevData = (TreeDataC *)prev->ViewData();
   genCount = GenerationCount(data->item);
   for (g=0; g<=genCount; g++) {

      GetDomain(data->item, g, &subDomain);

//
// If we're stacking vertically, see if our subdomain will fit horizontally
//   next to their tree
//
      if ( treeStyle == TREE_OUTLINE || rootGravity == WestGravity ) {
	 if ( prevData->domain.wd <= subDomain.wd ) {
	    *newDomain = data->domain;
	    int	y = data->domain.ymin - (subDomain.ymin - data->domain.ymin);
	    if ( y < marginHt ) y = marginHt;
	    newDomain->SetPos(data->domain.xmin, y);
	    return;
	 }
      }

//
// If we're stacking horizontally, see if our subdomain will fit vertically
//   next to their tree
//
      else if ( prevData->domain.ht <= subDomain.ht ) {
	 *newDomain = data->domain;
	 int	x = data->domain.xmin - (subDomain.xmin - data->domain.xmin);
	 if ( x < marginWd ) x = marginWd;
	 newDomain->SetPos(x, data->domain.ymin);
	 return;
      }

   } // End for each generation

} // End GetClosestFit

/*-----------------------------------------------------------------------
 *  Return the number of VISIBLE generations below a view item
 */

int
TreeViewC::GenerationCount(VItemC *item)
{
//
// Find the child with the largest generation count
//
   int		maxCount = -1; // This value will adjust for no visible children
   VItemC	*child   = item->child;
   TreeDataC	*data;
   while ( child ) {

      data = (TreeDataC *)child->ViewData();

      if ( DataVisible(data) ) {
	 int	count = GenerationCount(child);
	 if ( count > maxCount ) maxCount = count;
      }

      child = child->next;
   }

   return (1/*Our children*/ + maxCount/*Their descendants*/);

} // End GenerationCount

/*-----------------------------------------------------------------------
 *  Return the sub-domain for the specified generation.  Generation 0 is
 *     simply the bounding box of the item in question.
 */

void
TreeViewC::GetDomain(VItemC *item, int gen, RectC *domain)
{
   TreeDataC	*data = (TreeDataC *)item->ViewData();
   if ( !DataVisible(data) ) return;

   *domain = data->bounds;
   if ( gen == 0 || !item->child ) return;

   gen--;

   VItemC	*child   = item->child;
   TreeDataC	*childData;
   RectC	subDomain;
   while ( child ) {

//
// Process child only if visible
//
      childData = (TreeDataC *)child->ViewData();
      if ( DataVisible(childData) ) {

//
// Add the sub-domain of the child to the domain of the parent
//
	 GetDomain(child, gen, &subDomain);
	 *domain = *domain & subDomain;

      } // End if data valid and visible

      child = child->next;

   } // End for each item

} // End GetDomain

/*-----------------------------------------------------------------------
 *  Move the domain for the specified item.  This also affects the child
 *     domains and bounds.
 */

void
TreeViewC::SetDomainPos(TreeDataC *data, int x, int y)
{
   int	dx = x - data->domain.xmin;
   int	dy = y - data->domain.ymin;
   if ( dx == 0 && dy == 0 ) return;

   data->domain.SetPos(x, y);
   data->bounds.SetPos(data->bounds.xmin+dx, data->bounds.ymin+dy);

   VItemC	*child   = data->item->child;
   TreeDataC	*childData;

   while ( child ) {

      childData = (TreeDataC*)child->ViewData();
      if ( DataVisible(childData) )
	 SetDomainPos(childData, childData->domain.xmin+dx,
	 			 childData->domain.ymin+dy);

      child = child->next;
   }

} // End SetDomainPos

/*-----------------------------------------------------------------------
 *  Center the item with respect to the bounding boxes (not the domains)
 *     of its children.  The children must be centered first.
 */

void
TreeViewC::CenterItem(VItemC *item)
{
   TreeDataC	*data = (TreeDataC*)item->ViewData();
   if ( !item->child || !DataVisible(data) ) return;

   TreeDataC	*childData;
   int		valCount = 0;
   int		sum      = 0;
   VItemC	*child   = item->child;

//
// Average the child centers.
//
   while ( child ) {

      childData = (TreeDataC*)child->ViewData();
      if ( DataVisible(childData) ) {

//
// First center the child
//
	 CenterItem(child);

//
// Add bounds to averages
//
	 if ( rootGravity == NorthGravity )
	    sum += childData->bounds.xmin + childData->bounds.xmax;
	 else
	    sum += childData->bounds.ymin + childData->bounds.ymax;

	 valCount += 2;
      }

      child = child->next;
   }

   if ( valCount == 0 ) return;

//
// Position the item
//
   if ( rootGravity == NorthGravity ) {

      int	ix = (sum / valCount) - (data->bounds.wd / 2);
      data->bounds.SetPos(ix, data->bounds.ymin);

//
// Make sure the bounds haven't gone outside the domain
//
      int	dist = data->bounds.xmin - data->domain.xmin;
      if ( dist < 0 ) {
	 SetDomainPos(data, data->domain.xmin - dist, data->domain.ymin);
	 data->domain.SetSize(data->domain.wd + dist, data->domain.ht);
      }
   }

   else {
      int	iy = (sum / valCount) - (data->bounds.ht / 2);
      data->bounds.SetPos(data->bounds.xmin, iy);

//
// Make sure the bounds haven't gone outside the domain
//
      int	dist = data->bounds.ymin - data->domain.ymin;
      if ( dist < 0 ) {
	 SetDomainPos(data, data->domain.xmin, data->domain.ymin - dist);
	 data->domain.SetSize(data->domain.wd, data->domain.ht + dist);
      }
   }
   
} // End CenterItem

/*-----------------------------------------------------------------------
 *  Return size
 */

void
TreeViewC::GetSize(int *wd, int *ht)
{
   *wd = needWd;
   *ht = needHt;
}

/*-----------------------------------------------------------------------
 *  Redraw the items in this view
 */

void
TreeViewC::Redraw()
{
   Draw(viewBox->VisRect());
}

/*-----------------------------------------------------------------------
 *  Redraw the items in this view
 */

void
TreeViewC::Draw(RectC& area)
{
   if ( !viewBox->Realized() ) return;

   if ( needPlaces ) {
      PlaceItems();
      viewBox->UpdateScrollBars();
   }

//
// See how many items are visible
//
   VItemListC&	visItems = viewBox->VisItems();
   unsigned	count    = visItems.size();

//
// Highlight the first one if there is none highlighted
//
   if ( !hlItem || !visItems.includes(hlItem) )
      hlItem = count ? visItems[0] : NULL;

//
// Remove scrolling for ease of comparison
//
   RectC	sarea = area;
   sarea.xmin += viewBox->HScrollValue();
   sarea.xmax += viewBox->HScrollValue();
   sarea.ymin += viewBox->VScrollValue();
   sarea.ymax += viewBox->VScrollValue();

   GC		gc     = viewBox->ViewGC();
   Window	window = viewBox->ViewPm();

//
// Clear the drawing area
//
   XSetForeground(halApp->display, gc, bgColor);
   XFillRectangle(halApp->display, window, gc,
		  area.xmin, area.ymin, area.wd, area.ht);

//
// Loop through item list.
//
   for (int i=0; i<count; i++) {

      VItemC	*item = visItems[i];
      TreeDataC	*data = (TreeDataC *)item->ViewData();
      if ( !DataVisible(data) ) continue;

//
// Only draw this icon if it is at least partially visible
//
      if ( sarea.Overlaps(data->bounds) )
	 DrawItem(item, AS_IS, window);

//
// Draw the lines to the children if any part of the domain is visible
//
      if ( sarea.Overlaps(data->domain) )
	 DrawLines(item, window);

   } // End for each visible icon

   XCopyArea(halApp->display, window, viewBox->ViewWin(), gc, 0, 0,
	     viewBox->DAWidth(), viewBox->DAHeight(), 0, 0);

} // End Draw

/*-----------------------------------------------------------------------
 *  Public method to redraw the specified item
 */

void
TreeViewC::RedrawItem(const VItemC& item)
{
   if ( !viewBox->Realized() ) return;

//
// Get icon data
//
   TreeDataC	*data = (TreeDataC *)item.ViewData();
   if ( !DataVisible(data) ) return;

//
// Redraw if item not clipped out
//
   RectC	sarea = viewBox->VisRect();
   sarea.xmin += viewBox->HScrollValue();
   sarea.xmax += viewBox->HScrollValue();
   sarea.ymin += viewBox->VScrollValue();
   sarea.ymax += viewBox->VScrollValue();

   if ( sarea.Overlaps(data->bounds) )
      DrawItem((VItemC*)&item);

   if ( sarea.Overlaps(data->domain) )
      DrawLines((VItemC*)&item);

} // End RedrawItem

/*-----------------------------------------------------------------------
 *  Private method to redraw the specified item
 */

void
TreeViewC::DrawItem(VItemC *item, VItemDrawModeT mode, Drawable drawto)
{
   if ( !viewBox->Realized() ) return;
   TreeDataC	*data = (TreeDataC*)item->ViewData();
   if ( !DataVisible(data) ) return;

   Boolean	invert;

   if ( mode == AS_IS )
      invert = viewBox->SelItems().includes(item);
   else
      invert = (mode == INVERT);

//
// If drawto is NULL, draw to both the offscreen pixmap and the visible window
//
   Drawable	dst1 = drawto;
   Drawable	dst2 = (Drawable)NULL;
   if ( !dst1 ) {
      dst1 = viewBox->ViewWin();
      dst2 = viewBox->ViewPm();
   }
   GC		gc = viewBox->ViewGC();

//
// Figure out where on the screen the item should be drawn.  Take into
//    account scrolling and highlighting
//
   RectC	irect = data->bounds;
   irect.Set(data->bounds.xmin + hlThick - viewBox->HScrollValue(),
	     data->bounds.ymin + hlThick - viewBox->VScrollValue(),
	     data->bounds.wd - (hlThick*2),
	     data->bounds.ht - (hlThick*2));

   XSetForeground(halApp->display, gc, bgColor);
   XFillRectangle(halApp->display, dst1, gc,
		  irect.xmin, irect.ymin, irect.wd, irect.ht);
   item->Draw(irect, gc, dst1, invert);
   if ( dst2 ) {
      XSetForeground(halApp->display, gc, bgColor);
      XFillRectangle(halApp->display, dst2, gc,
		     irect.xmin, irect.ymin, irect.wd, irect.ht);
      item->Draw(irect, gc, dst1, invert);
   }

//
// Draw highlight if necessary
//
   if ( focusHere && item == hlItem )
      DrawHighlight(data, hlColor, drawto);

} // End DrawItem

/*-----------------------------------------------------------------------
 *  Private method to draw the lines from the specified item to its
 *     children
 */

void
TreeViewC::DrawLines(VItemC *item, Drawable drawto)
{
   if ( !viewBox->Realized() ) return;

   TreeDataC	*data = (TreeDataC*)item->ViewData();
   if ( !DataVisible(data) ) return;

   VItemC	*child = item->child;
   if ( !child ) return;

//
// Draw lines connecting children
//
   int		x0, y0;
   int		x1, y1;
   TreeDataC	*childData;
   Boolean	childDrawn = False;

   if ( treeStyle == TREE_OUTLINE ) {

      x0 = x1 = (data->bounds.xmin + data->bounds.xmax) / 2;
      y0 = y1 = data->bounds.ymax;

      while ( child ) {

	 childData = (TreeDataC*)child->ViewData();
	 if ( DataVisible(childData) ) {

//
// Draw line across to child
//
	    x1 = childData->bounds.xmin;
	    y1 = (childData->bounds.ymin + childData->bounds.ymax) / 2;
	    DrawLine(x0, y1, x1, y1, drawto);
	    childDrawn = True;
	 }

	 child = child->next;
      }

//
// Draw line down from bottom of item to middle of last child
//
      if ( childDrawn )
	 DrawLine(x0, y0, x0, y1, drawto);
	 
   } // End if outline

   else if ( treeStyle == TREE_ANGLED ) {

      if ( rootGravity == NorthGravity ) {

	 x0 = (data->bounds.xmin + data->bounds.xmax) / 2;
	 y0 = data->bounds.ymax;

	 while ( child ) {

	    childData = (TreeDataC*)child->ViewData();
	    if ( DataVisible(childData) ) {

//
// Draw line across to child
//
	       x1 = (childData->bounds.xmin + childData->bounds.xmax) / 2;
	       y1 = childData->bounds.ymin;

//
// Improve lines that are slightly off vertical
//
	       int	dist = x1 - x0;
	       if ( dist < 0 ) dist = -dist;
	       if ( dist == 1 ) x1 = x0;

	       DrawLine(x0, y0, x1, y1, drawto);
	       childDrawn = True;
	    }

	    child = child->next;
	 }

      } // End if North gravity

      else {

	 x0 = data->bounds.xmax;
	 y0 = (data->bounds.ymin + data->bounds.ymax) / 2;

	 while ( child ) {

	    childData = (TreeDataC*)child->ViewData();
	    if ( DataVisible(childData) ) {

//
// Draw line across to child
//
	       x1 = childData->bounds.xmin;
	       y1 = (childData->bounds.ymin + childData->bounds.ymax) / 2;

//
// Improve lines that are slightly off horizontal
//
	       int	dist = y1 - y0;
	       if ( dist < 0 ) dist = -dist;
	       if ( dist == 1 ) y1 = y0;

	       DrawLine(x0, y0, x1, y1, drawto);
	       childDrawn = True;
	    }

	    child = child->next;
	 }

      } // End if West gravity

   } // End if angled lines

   else { // TREE_SQUARED

      if ( rootGravity == NorthGravity ) {

//
// Calculate midpoint between parent and children
//
	 x0 = (data->bounds.xmin + data->bounds.xmax) / 2;
	 y0 = data->bounds.ymax;
	 int	ymid = data->bounds.ymax + (genSpacing / 2);
	 int	xpar = x0;

//
// Loop through children
//
	 int	xmin = -1;
	 int	xmax = -1;

	 while ( child ) {

	    childData = (TreeDataC*)child->ViewData();
	    if ( DataVisible(childData) ) {

//
// Draw line from midpoint to child
//
	       x1 = (childData->bounds.xmin + childData->bounds.xmax) / 2;
	       y1 = childData->bounds.ymin;

//
// Kludge for lines that are slightly off center
//
	       int	dist = x1 - xpar;
	       if ( dist < 0 ) dist = -dist;
	       if ( dist == 1 ) x1 = xpar;

	       DrawLine(x1, ymid, x1, y1, drawto);

	       if ( xmin < 0  ) xmin = x1;
	       if ( x1 > xmax ) xmax = x1;

	       childDrawn = True;
	    }

	    child = child->next;
	 }

//
// Draw horizontal midpoint line
//
	 if ( childDrawn ) {
	    DrawLine(x0, y0, x0, ymid, drawto);
	    DrawLine(xmin, ymid, xmax, ymid, drawto);
	 }

      } // End if North gravity

      else {

//
// Calculate midpoint between parent and children
//
	 x0 = data->bounds.xmax;
	 y0 = (data->bounds.ymin + data->bounds.ymax) / 2;
	 int	xmid = data->bounds.xmax + (genSpacing / 2);
	 int	ypar = y0;

//
// Loop through children
//
	 int	ymin = -1;
	 int	ymax = -1;

	 while ( child ) {

	    childData = (TreeDataC*)child->ViewData();
	    if ( DataVisible(childData) ) {

//
// Draw line from midpoint to child
//
	       x1 = childData->bounds.xmin;
	       y1 = (childData->bounds.ymin + childData->bounds.ymax) / 2;

//
// Kludge for lines that are slightly off center
//
	       int	dist = y1 - ypar;
	       if ( dist < 0 ) dist = -dist;
	       if ( dist == 1 ) y1 = ypar;

	       DrawLine(xmid, y1, x1, y1, drawto);

	       if ( ymin < 0  ) ymin = y1;
	       if ( y1 > ymax ) ymax = y1;

	       childDrawn = True;
	    }

	    child = child->next;
	 }

//
// Draw vertical midpoint line
//
	 if ( childDrawn ) {
	    DrawLine(x0, y0, xmid, y0, drawto);
	    DrawLine(xmid, ymin, xmid, ymax, drawto);
	 }

      } // End if West gravity

   } // End if squared lines

} // End DrawLines

/*-----------------------------------------------------------------------
 *  Method to draw the specified line using the current line style and width
 */

void
TreeViewC::DrawLine(int x0, int y0, int x1, int y1, Drawable drawto)
{
//
// Scroll points into drawing space
//
   x0 -= viewBox->HScrollValue();
   x1 -= viewBox->HScrollValue();
   y0 -= viewBox->VScrollValue();
   y1 -= viewBox->VScrollValue();

//
// Throw out lines that don't cross visible space
//
   int	xmax = needWd - 1;
   int	ymax = needHt - 1;
   if ( (x0 <= 0    && x1 <= 0   ) ||
        (x0 >= xmax && x1 >= xmax) ||
	(y0 <= 0    && y1 <= 0   ) ||
        (y0 >= ymax && y1 >= ymax) ) return;

//
// Make sure x0 is left of x1
//
   int	tmp;
   if ( x0 > x1 ) {
      tmp = x0; x0 = x1; x1 = tmp;
      tmp = y0; y0 = y1; y1 = tmp;
   }

//
// Calculate slope of line
//
   int		dx    = (int)(x1 - x0);
   int		dy    = (int)(y1 - y0);
   float	slope = (dx != 0) ? (float)dy/(float)dx : 0;
   int		x2, y2;

   if ( dx != 0 ) {

//
// Clip to left side of drawing area
//
      if ( x0 < 0 ) {
	 x2 = 0;
	 y2 = (int)(y1 - (slope * (x1 - x2)));
	 x0 = x2;
	 y0 = y2;
      }

//
// Clip to right side of drawing area
//
      if ( x1 > xmax ) {
	 x2 = xmax;
	 y2 = (int)(y1 - (slope * (x1 - x2)));
	 x1 = x2;
	 y1 = y2;
      }

//
// Throw out lines that don't cross visible space
//
      if ( (y0 <= 0    && y1 <= 0   ) ||
	   (y0 >= ymax && y1 >= ymax) )
	 return;

   } // End if line is not vertical

//
// Make sure y0 is above of y1
//
   if ( y0 > y1 ) {
      tmp = x0; x0 = x1; x1 = tmp;
      tmp = y0; y0 = y1; y1 = tmp;
      dx    = (int)(x1 - x0);
      dy    = (int)(y1 - y0);
      if ( dx != 0 ) slope = (float)dy/(float)dx;
      else	     slope = 0.0;
   }

   if ( dx == 0 ) {
      if ( y0 < 0    ) y0 = 0;
      if ( y1 > ymax ) y1 = ymax;
   }

   else {

//
// Clip to top side of drawing area
//
      if ( y0 < 0 ) {
	 y2 = 0;
	 x2 = (int)(x1 - ((y1 - y2) / slope));
	 y0 = y2;
	 x0 = x2;
      }

//
// Clip to bottom side of drawing area
//
      if ( y1 > ymax ) {
	 y2 = ymax;
	 x2 = (int)(x1 - ((y1 - y2) / slope));
	 y1 = y2;
	 x1 = x2;
      }

   } // End if line is not vertical

//
// If drawto is NULL, draw to both the offscreen pixmap and the visible window
//
   Drawable	dst1 = drawto;
   Drawable	dst2 = (Drawable)NULL;
   if ( !dst1 ) {
      dst1 = viewBox->ViewWin();
      dst2 = viewBox->ViewPm();
   }
   GC		gc = viewBox->ViewGC();

//
// Draw lines
//
   int	lstyle = LineSolid;
   if ( lineStyle == XmSHADOW_ETCHED_IN_DASH ||
        lineStyle == XmSHADOW_ETCHED_OUT_DASH ||
	lineStyle == XmSINGLE_DASHED_LINE ) lstyle = LineOnOffDash;

   XSetLineAttributes(halApp->display, gc, lineWidth, lstyle, CapButt,
		      JoinBevel);

   if ( lineStyle == XmSHADOW_ETCHED_OUT ||
	lineStyle == XmSHADOW_ETCHED_OUT_DASH )
      XSetForeground(halApp->display, gc, etchColor);
   else
      XSetForeground(halApp->display, gc, lineColor);

   if ( debuglev > 1 )
      cout <<"Drawing line: " <<x0 <<" " <<y0 <<" to "
			      <<x1 <<" " <<y1 <<endl;

   XDrawLine(halApp->display, dst1, gc, x0, y0, x1, y1);
   if ( dst2 ) XDrawLine(halApp->display, dst2, gc, x0, y0, x1, y1);

   if ( lineStyle == XmSHADOW_ETCHED_OUT ||
	lineStyle == XmSHADOW_ETCHED_OUT_DASH )
      XSetForeground(halApp->display, gc, lineColor);
   else
      XSetForeground(halApp->display, gc, etchColor);

   if ( lineStyle == XmSHADOW_ETCHED_IN ||
	lineStyle == XmSHADOW_ETCHED_IN_DASH ||
	lineStyle == XmSHADOW_ETCHED_OUT ||
	lineStyle == XmSHADOW_ETCHED_OUT_DASH ) {
      int	off = lineWidth > 0 ? lineWidth : 1;
      x0 += off;
      x1 += off;
      y0 += off;
      y1 += off;
      XDrawLine(halApp->display, dst1, gc, x0, y0, x1, y1);
      if ( dst2 ) XDrawLine(halApp->display, dst2, gc, x0, y0, x1, y1);
   }

} // End DrawLine

/*-----------------------------------------------------------------------
 *  Find the item that contains the given coordinate
 */

VItemC*
TreeViewC::PickItem(int x, int y)
{
//
// Allow a 10-pixel buffer around the pick.
//
   int		xSpacing = 10;
   int		ySpacing = 10;
   RectC	rect(x-xSpacing, y-ySpacing, xSpacing*2, ySpacing*2);
   VItemListC	list;
   PickItems(rect, list);

   VItemC	*item = NULL;

//
// If there is more than 1 item, pick the closest
//
   if ( list.size() == 1 )

      item = list[0];

   else if ( list.size() > 1 ) {

      int	minDist = MAXINT;
      unsigned	count = list.size();
      for (int i=0; i<count; i++) {

	 VItemC		*li = list[i];
	 TreeDataC	*data = (TreeDataC *)li->ViewData();

	 if ( data->bounds.Inside(x, y) ) {
	    item = li;
	    i = count;
	 }

	 else {

	    int	midX = data->bounds.xmin + (data->bounds.wd/2);
	    int	midY = data->bounds.ymin + (data->bounds.ht/2);
	    int	dx = x - midX;
	    int	dy = y - midY;
	    dx *= dx;
	    dy *= dy;

	    int	dist = (int)sqrt((double)(dx + dy));
	    if ( dist <= minDist ) {
	       minDist = dist;
	       item = li;
	    }
	 }

      } // End for each picked item

   } // End if more than one item picked

   return item;

} // End PickItem

/*-----------------------------------------------------------------------
 *  Return the items in the pick rect
 */

void
TreeViewC::PickItems(RectC &pickRect, VItemListC& list)
{
   VItemListC&	visItems = viewBox->VisItems();
   unsigned	count = visItems.size();
   for (int i=0; i<count; i++) {

      VItemC	*item = visItems[i];
      TreeDataC *data = (TreeDataC *)item->ViewData();
      if ( !DataVisible(data) ) continue;

//
// See if the item area overlaps the pick rectangle
//
      if ( data->bounds.Overlaps(pickRect) ) list.append(item);

   } // End for each visible item

} // End PickItems

/*-----------------------------------------------------------------------
 *  Change the visibility of an item
 */

void
TreeViewC::ChangeItemVis(const VItemC& item)
{
   needPlaces = True;

   TreeDataC	*data = (TreeDataC *)item.ViewData();
   if ( DataValid(data) )
      data->visible = viewBox->VisItems().includes(&item);

//
// Redraw if necessary
//
   if ( !viewBox->Deferred() ) Redraw();
   else			       viewBox->Changed(True);

} // End ChangeItemVis

/*-----------------------------------------------------------------------
 *  Change the visibility of unknown items
 */

void
TreeViewC::ChangeItemVis()
{
   needPlaces = True;

//
// Redraw if necessary
//
   if ( !viewBox->Deferred() ) Redraw();
   else			       viewBox->Changed(True);

} // End ChangeItemVis

/*-----------------------------------------------------------------------
 *  Flash an item
 */

void
TreeViewC::FlashItem(const VItemC *item)
{
//
// If an item is already flashing, stop it
//
   if ( flashTimer ) {
      XtRemoveTimeOut(flashTimer);
      RedrawItem(*flashItem);
   }

//
// Make sure the item is visible
//
   ScrollToItem(*item);

//
// Start a time out proc to flash the item a few times
//
   flashItem  = (VItemC *)item;
   flashCount = viewBox->FindBlinkCount();
   flashOn    = viewBox->SelItems().includes(item);
   flashTimer = XtAppAddTimeOut(halApp->context, viewBox->FindBlinkInterval(),
				(XtTimerCallbackProc)FlashProc,
			        (XtPointer)this);

} // End FlashItem

/*-----------------------------------------------------------------------
 * Scroll the view window so the specified item is visible
 */

void
TreeViewC::ScrollToItem(const VItemC& item)
{
   TreeDataC	*data = (TreeDataC *)item.ViewData();
   if ( DataVisible(data) ) ScrollToItem(data);
}

void
TreeViewC::ScrollToItem(TreeDataC *data)
{
   if ( !data ) return;

   int	val, size, inc, pinc, min, max;

   int	ymax = viewBox->VisRect().ymax + viewBox->VScrollValue();
   int	ymin = viewBox->VisRect().ymin + viewBox->VScrollValue();

//
// Check top and bottom
//
   if ( data->bounds.ymax > ymax || viewBox->VisRect().ht < data->bounds.ht ) {

//
// It is off the bottom, so position the bottom of the scroll bar.
//
      XmScrollBarGetValues(viewBox->VScrollBar(), &val, &size, &inc, &pinc);
      XtVaGetValues(viewBox->VScrollBar(), XmNmaximum, &max, NULL);

      val = data->bounds.ymax;
      if ( treeStyle == TREE_OUTLINE || rootGravity == WestGravity )
	 val += sibSpacing;
      else
	 val += genSpacing;
      if ( val > max ) val = max;
      val -= size;

      XmScrollBarSetValues(viewBox->VScrollBar(), val, size, inc, pinc,
			   /*notify*/True);
   }

   else if ( data->bounds.ymin < ymin ) {

//
// It is off the top, so position the top of the scroll bar.
//
      XmScrollBarGetValues(viewBox->VScrollBar(), &val, &size, &inc, &pinc);
      XtVaGetValues(viewBox->VScrollBar(), XmNminimum, &min, NULL);

      val = data->bounds.ymin;
      if ( treeStyle == TREE_OUTLINE || rootGravity == WestGravity )
	 val -= sibSpacing;
      else
	 val -= genSpacing;
      if ( val < min ) val = min;

      XmScrollBarSetValues(viewBox->VScrollBar(), val, size, inc, pinc,
			   /*notify*/True);
   }

//
// Check left and right
//
   int	xmax = viewBox->VisRect().xmax + viewBox->HScrollValue();
   int	xmin = viewBox->VisRect().xmin + viewBox->HScrollValue();

   if ( data->bounds.xmax > xmax || viewBox->VisRect().wd < data->bounds.wd ) {

//
// It is off the right, so position the right of the scroll bar.
//
      XmScrollBarGetValues(viewBox->HScrollBar(), &val, &size, &inc, &pinc);
      XtVaGetValues(viewBox->HScrollBar(), XmNmaximum, &max, NULL);

      val = data->bounds.xmax;
      if ( treeStyle == TREE_OUTLINE || rootGravity == WestGravity )
	 val += genSpacing;
      else
	 val += sibSpacing;
      if ( val > max ) val = max;
      val -= size;

      XmScrollBarSetValues(viewBox->HScrollBar(), val, size, inc, pinc,
			   /*notify*/True);
   }

   else if ( data->bounds.xmin < xmin ) {

//
// It is off the left, so position the left of the scroll bar.
//
      XmScrollBarGetValues(viewBox->HScrollBar(), &val, &size, &inc, &pinc);
      XtVaGetValues(viewBox->HScrollBar(), XmNminimum, &min, NULL);

      val = data->bounds.xmin;
      if ( treeStyle == TREE_OUTLINE || rootGravity == WestGravity )
	 val -= genSpacing;
      else
	 val -= sibSpacing;
      if ( val < min ) val = min;

      XmScrollBarSetValues(viewBox->HScrollBar(), val, size, inc, pinc,
			   /*notify*/True);
   }

} // End ScrollToItem

/*-----------------------------------------------------------------------
 *  Timer proc used to flash an item
 */

void
TreeViewC::FlashProc(TreeViewC *This, XtIntervalId*)
{
   if ( This->flashCount == 0 ) {
      This->RedrawItem(*This->flashItem);
      This->flashItem = NULL;
      This->flashTimer = (XtIntervalId)NULL;
      return;
   }

//
// Flip the state of the blinking item
//
   if ( This->flashOn ) This->DrawItem(This->flashItem, NORMAL);
   else                 This->DrawItem(This->flashItem, INVERT);
   This->flashOn = !This->flashOn;
   This->flashCount--;

//
// Add this timeout again
//
   This->flashTimer =
      XtAppAddTimeOut(halApp->context, This->viewBox->FindBlinkInterval(),
		      (XtTimerCallbackProc)FlashProc, (XtPointer)This);

} // End FlashProc

/*-----------------------------------------------------------------------
 *  Draw a highlight border around an item using the given color
 */

void
TreeViewC::DrawHighlight(VItemC *item, Pixel color)
{
   if ( !item ) return;

   TreeDataC	*data = (TreeDataC *)item->ViewData();
   if ( DataVisible(data) ) DrawHighlight(data, color);
}

void
TreeViewC::DrawHighlight(TreeDataC *data, Pixel color, Drawable drawto)
{
   if ( !data ) return;

   XRectangle	rects[4];
   XRectangle	*rect = rects;

   int		xoff = viewBox->HScrollValue();
   int		yoff = viewBox->VScrollValue();

   rect->x      = data->bounds.xmin - xoff;
   rect->y      = data->bounds.ymin - yoff;
   rect->width  = data->bounds.wd;
   rect->height = hlThick;

   rect++;
   rect->x      = data->bounds.xmin - xoff;
   rect->y      = data->bounds.ymin - yoff;
   rect->width  = hlThick;
   rect->height = data->bounds.ht;

   rect++;
   rect->x      = data->bounds.xmax - hlThick + 1 - xoff;
   rect->y      = data->bounds.ymin - yoff;
   rect->width  = hlThick;
   rect->height = data->bounds.ht;

   rect++;
   rect->x      = data->bounds.xmin - xoff;
   rect->y      = data->bounds.ymax - hlThick + 1 - yoff;
   rect->width  = data->bounds.wd;
   rect->height = hlThick;

   GC	gc = viewBox->ViewGC();
   XSetForeground(halApp->display, gc, color);

//
// If drawto is NULL, draw to both the offscreen pixmap and the visible window
//
   Drawable	dst1 = drawto;
   Drawable	dst2 = (Drawable)NULL;
   if ( !dst1 ) {
      dst1 = viewBox->ViewWin();
      dst2 = viewBox->ViewPm();
   }

   XFillRectangles(halApp->display, dst1, gc, rects, 4);
   if ( dst2 )
      XFillRectangles(halApp->display, dst2, gc, rects, 4);

} // End DrawHighlight

/*-----------------------------------------------------------------------
 *  Handle a keyboard focus change event in the area
 */

void
TreeViewC::HandleFocusChange(Widget, TreeViewC *This, XEvent *ev, Boolean *)
{
   if ( !This->shown ) return;

   switch (ev->type) {

      case (FocusIn):
      case (EnterNotify):
	 This->focusHere = True;
	 This->DrawHighlight(This->hlItem, This->hlColor);
	 break;

      case (FocusOut):
      case (LeaveNotify):
	 This->focusHere = False;
	 This->DrawHighlight(This->hlItem, This->bgColor);
	 break;

      default:
	 break;
   }

} // End HandleFocusChange

/*-----------------------------------------------------------------------
 *  Handle an input event
 */

void
TreeViewC::HandleKeyPress(XKeyEvent *ev)
{
   if ( !focusHere || !hlItem ) return;

   TreeDataC	*data = (TreeDataC *)hlItem->ViewData();
   if ( !DataVisible(data) ) return;

   KeySym	keysym;
   Modifiers	state;
   XtTranslateKey(halApp->display, ev->keycode, ev->state, &state, &keysym);

   VItemListC&	visItems  = viewBox->VisItems();
   unsigned	size      = visItems.size();
   VItemC	*newItem = NULL;

   switch (keysym) {

      case (XK_Home):
	 newItem = visItems[0];
	 break;

      case (XK_End):
	 newItem = visItems[size-1];
	 break;

//
// Left means:
//
// West gravity  - parent
// North gravity - previous sibling
//
// Ctrl-Left means:
//
// West gravity  - root
// North gravity - first child of parent
//
      case (XK_Left):
	 if ( ev->state & ControlMask ) {
	    if ( treeStyle == TREE_OUTLINE || rootGravity == WestGravity )
	       newItem = RootOf(hlItem);
	    else
	       newItem = FirstVisibleChild(hlItem->parent);
	 }
	 else if ( treeStyle == TREE_OUTLINE || rootGravity == WestGravity )
	    newItem = hlItem->parent;
	 else
	    newItem = hlItem->prev;
	 break;

//
// Right means:
//
// West gravity  - first child
// North gravity - next sibling
//
// Ctrl-Right means:
//
// West gravity  - last child
// North gravity - last child of parent
//
      case (XK_Right):
	 if ( ev->state & ControlMask ) {
	    if ( treeStyle == TREE_OUTLINE || rootGravity == WestGravity )
	       newItem = LastVisibleChild(hlItem);
	    else
	       newItem = LastVisibleChild(hlItem->parent);
	 }
	 else if ( treeStyle == TREE_OUTLINE || rootGravity == WestGravity )
	    newItem = hlItem->child;
	 else
	    newItem = hlItem->next;
	 break;

//
// Up means:
//
// West gravity  - previous sibling
// North gravity - parent
//
// Ctrl-Up means:
//
// West gravity  - first child of parent
// North gravity - root
//
      case (XK_Up):
	 if ( ev->state & ControlMask ) {
	    if ( treeStyle == TREE_OUTLINE || rootGravity == WestGravity )
	       newItem = FirstVisibleChild(hlItem->parent);
	    else
	       newItem = RootOf(hlItem);
	 }
	 else if ( treeStyle == TREE_OUTLINE || rootGravity == WestGravity )
	    newItem = hlItem->prev;
	 else
	    newItem = hlItem->parent;
	 break;

//
// Down means:
//
// West gravity  - next sibling
// North gravity - first child
//
// Ctrl-Down means:
//
// West gravity  - last child of parent
// North gravity - last child
//
      case (XK_Down):
	 if ( ev->state & ControlMask ) {
	    if ( treeStyle == TREE_OUTLINE || rootGravity == WestGravity )
	       newItem = LastVisibleChild(hlItem->parent);
	    else
	       newItem = LastVisibleChild(hlItem);
	 }
	 else if ( treeStyle == TREE_OUTLINE || rootGravity == WestGravity )
	    newItem = hlItem->next;
	 else
	    newItem = hlItem->child;
	 break;

#if 0
      case (XK_Page_Up): {
	 char	*parms = "0";
	 XtCallActionProc(viewBox->VScrollBar(), "PageUpOrLeft", (XEvent*)ev,
	 		  &parms, 1);
      } break;

      case (XK_Page_Down): {
	 char	*parms = "0";
	 XtCallActionProc(viewBox->VScrollBar(), "PageDownOrRight", (XEvent*)ev,
	 		  &parms, 1);
      } break;
#endif

      case (XK_space):
	 if ( ev->state & (ShiftMask|ControlMask) )
	    viewBox->ToggleItem(*hlItem);
	 else
	    viewBox->SelectItemOnly(*hlItem);
	 break;

   } // End switch keysym

   if ( newItem )
      HighlightItem(newItem);

} // End HandleKeyPress

/*-----------------------------------------------------------------------
 *  Handle press of mouse button 1
 */

void
TreeViewC::HandleButton1Press(XButtonEvent *be)
{
//
// There is a chance that several different selections/deselections will take
//    place.  Wait for the button release to really do them.
//
   viewBox->PartialSelect(True);

//
// See if this is part of a multiple click
//
   unsigned long	elapsed = be->time - lastClickTime;
   if ( elapsed <= XtGetMultiClickTime(halApp->display) )
      clickCount++;
   else
      clickCount = 1;
   lastClickTime = be->time;

//
// Translate the position by the scroll factors
//
   pickX = be->x + viewBox->HScrollValue();
   pickY = be->y + viewBox->VScrollValue();

//
// Initialize the rubberband rectangle
//
   pickRect.Set(pickX, pickY, 1, 1);

//
// Pick and highlight press item
//
   pickList.removeAll();
   pressItem = PickItem(pickX, pickY);

   pickDrawMode = INVERT;
   pickState = be->state;
   if ( (pickState & ControlMask) == ControlMask ) {
      if ( pressItem && viewBox->SelItems().includes(pressItem) )
	 pickDrawMode = NORMAL;
   }

   else if ( (pickState & ShiftMask) == ShiftMask ) {
      ;
   }

   else if ( clickCount == 1 ) {
      viewBox->DeselectAllItems();
      if ( pressItem ) DrawItem(pressItem, pickDrawMode);
   }

   if ( pressItem ) {
//      DrawItem(*pressItem, pickDrawMode);
      pickList.add(pressItem);
      HighlightItem(pressItem);
   }

   XDrawRectangle(halApp->display, viewBox->ViewWin(), pickGC,
   		  pickRect.xmin - viewBox->HScrollValue(),
		  pickRect.ymin - viewBox->VScrollValue(),
		  pickRect.wd, pickRect.ht);

} // End HandleButton1Press

/*-----------------------------------------------------------------------
 *  Handle release of mouse button 1
 */

void
TreeViewC::HandleButton1Release(XButtonEvent *be)
{
   if ( scrollTimer ) XtRemoveTimeOut(scrollTimer);

//
// Clear the current rectangle
//
   XDrawRectangle(halApp->display, viewBox->ViewWin(), pickGC,
   		  pickRect.xmin - viewBox->HScrollValue(),
		  pickRect.ymin - viewBox->VScrollValue(),
		  pickRect.wd, pickRect.ht);

//
// Get the list of items in the rectangle
//
   PickItems(pickRect, pickList);

//
// Perform single or double click actions.  If the single click fails,
//    reset the lastClickTime so we don't get a double click
//
   if ( clickCount == 1 )
      HandleSingleClick(be);
   else if ( clickCount == 2 )
      HandleDoubleClick(be);

//
// Now finish processing any selections that were queued up.
//
   viewBox->PartialSelect(False);

} // End HandleButton1Release

/*-----------------------------------------------------------------------
 *  Handle a single click event in the area.
 */

void
TreeViewC::HandleSingleClick(XButtonEvent *be)
{
//
// If there isn't exactly one item selected, the the lastClickTime to 0 so
//    we don't get a double click
//
   if ( pickList.size() != 1 ) lastClickTime = 0;

//
// If no items are selected, return
//
   if ( pickList.size() < 1 ) return;

//
// See how long this click lasted
//
   int	clickLength = (int)(be->time - lastClickTime);

//
// If there is a single item and this is a quick click,
//    give the item a chance to process the event.
//
   if ( pickList.size() == 1 &&
        clickLength <= XtGetMultiClickTime(halApp->display) ) {

      TreeDataC	*data = (TreeDataC*)pressItem->ViewData();
      if ( DataValid(data) ) {

//
// Translate the position by the scroll factors and the item bounds
//
	 int	x = be->x + viewBox->HScrollValue() - data->bounds.xmin;
	 int	y = be->y + viewBox->VScrollValue() - data->bounds.ymin;

	 if ( pressItem->SingleClick(x, y, pickState) ) return;
      }
   }

//
// Update the selection
//
   if ( (pickState & ShiftMask) == ShiftMask )
      viewBox->SelectItems(pickList);
   else if ( (pickState & ControlMask) == ControlMask ) {
      if ( pickDrawMode == INVERT ) // Selecting
	 viewBox->SelectItems(pickList);
      else
	 viewBox->DeselectItems(pickList);
   }
   else
      viewBox->SelectItemsOnly(pickList);

} // End HandleSingleClick

/*-----------------------------------------------------------------------
 *  Handle a double click event in the area
 */

void
TreeViewC::HandleDoubleClick(XButtonEvent *be)
{
   if ( !pressItem ) return;

//
// Translate the position by the scroll factors
//
   int	x = be->x + viewBox->HScrollValue();
   int	y = be->y + viewBox->VScrollValue();

//
// See which icon is under pointer.  If this is not the same item as the
//    press event, start over
//
   VItemC	*releaseItem = PickItem(x, y);
   if ( releaseItem != pressItem ) return;

//
// Give the item a chance to process the event.
//
   TreeDataC	*data = (TreeDataC*)pressItem->ViewData();
   if ( DataValid(data) &&
        pressItem->DoubleClick(x-data->bounds.xmin,
			       y-data->bounds.ymin, pickState) ) return;

//
// Highlight the item
//
   HighlightItem(pressItem);

//
// Open the item
//
   pressItem->CallOpenCallbacks();
 
} // End HandleDoubleClick

/*-----------------------------------------------------------------------
 *  Handle pointer motion with button 1 pressed
 */

void
TreeViewC::HandleButton1Motion(XMotionEvent *ev)
{
//
// Clear the current rectangle
//
   XDrawRectangle(halApp->display, viewBox->ViewWin(), pickGC,
		  pickRect.xmin - viewBox->HScrollValue(),
		  pickRect.ymin - viewBox->VScrollValue(),
		  pickRect.wd, pickRect.ht);

//
// If the pointer is outside the bounds, start a timer to auto-scroll
//
   Boolean	off_top = (ev->y <= 0);
   Boolean	off_bot = (ev->y >= (int)viewBox->DAHeight());
   if ( off_top || off_bot ) {

//
// Return if no scrolling is possible
//
      if ( !viewBox->VScrollOn() ||
	   (off_top && viewBox->VScrollValue() == 0) ||
	   (off_bot && viewBox->VScrollValue() == viewBox->VScrollMaxValue()) )
	 return;

      scrollEvent = *ev;

//
// Start scrolling if we're not already
//
      if ( !scrollTimer ) {

//                               
// Call the action proc to scroll
//                                                         
         char   *parms = "0";                              
	 scrollAction = off_top ? "IncrementUpOrLeft" : "IncrementDownOrRight";
	 XtCallActionProc(viewBox->VScrollBar(), scrollAction,
			  (XEvent*)&scrollEvent, &parms, 1);

//
// Start automatic scrolling
//
	 scrollTimer  = XtAppAddTimeOut(halApp->context, 250,
					(XtTimerCallbackProc)HandleAutoScroll,
					(XtPointer)this);

      } // End if not already scrolling

   } // End if pointer hit top or bottom of window

   else {

//
// Stop any auto scrolling
//
      if ( scrollTimer ) XtRemoveTimeOut(scrollTimer);
      scrollTimer = (XtIntervalId)NULL;
   }

//
// Translate the position by the scroll factors
//
   int	x = ev->x + viewBox->HScrollValue();
   int	y = ev->y + viewBox->VScrollValue();

   UpdatePickRect(x, y);

} // End HandleButton1Motion

/*-----------------------------------------------------------------------
 *  Handle automatic scrolling
 */

void
TreeViewC::HandleAutoScroll(TreeViewC *This, XtIntervalId*)
{
//
// Clear the current rectangle
//
   XDrawRectangle(halApp->display, This->viewBox->ViewWin(), This->pickGC,
		  This->pickRect.xmin - This->viewBox->HScrollValue(),
		  This->pickRect.ymin - This->viewBox->VScrollValue(),
		  This->pickRect.wd, This->pickRect.ht);

//
// Call the action proc to scroll
//
   char	*parms = "0";
   XtCallActionProc(This->viewBox->VScrollBar(), This->scrollAction,
		    (XEvent*)&This->scrollEvent, &parms, 1);

//
// Redraw the items in the pick list since the scroll cleared them
//
   RectC	sarea = This->viewBox->VisRect();
   sarea.xmin += This->viewBox->HScrollValue();
   sarea.xmax += This->viewBox->HScrollValue();
   sarea.ymin += This->viewBox->VScrollValue();
   sarea.ymax += This->viewBox->VScrollValue();

   unsigned	count = This->pickList.size();
   for (int i=0; i<count; i++) {

      VItemC	*item = This->pickList[i];
      TreeDataC	*data = (TreeDataC *)item->ViewData();

//
// Redraw if item not clipped out
//
      if ( sarea.Overlaps(data->bounds) )
	 This->DrawItem(item, This->pickDrawMode);
   }

//
// Update the rectangle
//
   int	x = This->scrollEvent.x + This->viewBox->HScrollValue();
   int	y = This->scrollEvent.y + This->viewBox->VScrollValue();
   This->UpdatePickRect(x, y);

//
// Repeat the cycle
//
   This->scrollTimer = XtAppAddTimeOut(halApp->context, 250,
				       (XtTimerCallbackProc)HandleAutoScroll,
				       (XtPointer)This);

} // End HandleAutoScroll

/*-----------------------------------------------------------------------
 *  Redraw the pick rectangle
 */

void
TreeViewC::UpdatePickRect(int x, int y)
{
//
// Update the pick rectangle
//
   if ( pickX > x ) {
      pickRect.xmin = x;
      pickRect.xmax = pickX;
   } else {
      pickRect.xmin = pickX;
      pickRect.xmax = x;
   }

   if ( pickY > y ) {
      pickRect.ymin = y;
      pickRect.ymax = pickY;
   } else {
      pickRect.ymin = pickY;
      pickRect.ymax = y;
   }

   pickRect.wd = pickRect.xmax - pickRect.xmin + 1;
   pickRect.ht = pickRect.ymax - pickRect.ymin + 1;

//
// Get a new list of items in the pick rectangle
//
   VItemListC	list;
   PickItems(pickRect, list);

//
// Remove items in old list, but not in new list 
//
   unsigned	count = pickList.size();
   for (int i=count-1; i>=0; i--) {
      VItemC	*item = pickList[i];
      if ( !list.includes(item) ) {
	 DrawItem(item, AS_IS);
	 pickList.remove(i);
      }
   }

//
// Add items in new list, but not in old list 
//
   count = list.size();
   for (i=0; i<count; i++) {
      VItemC	*item = list[i];
      if ( !pickList.includes(item) ) {
	 DrawItem(item, pickDrawMode);
	 pickList.add(item);
      }
   }

//
// Draw the new rectangle
//
   XDrawRectangle(halApp->display, viewBox->ViewWin(), pickGC,
		  pickRect.xmin - viewBox->HScrollValue(),
		  pickRect.ymin - viewBox->VScrollValue(),
		  pickRect.wd, pickRect.ht);

} // End UpdatePickRect

/*-----------------------------------------------------------------------
 *  Highlight an item
 */

void
TreeViewC::HighlightItem(const VItemC *item)
{
   if ( item == hlItem ) return;

//
// Remove current highlight
//
   if ( hlItem ) {
      DrawHighlight(hlItem, bgColor);
   }
   
   hlItem = (VItemC *)item;
   if ( !hlItem ) return;

//
// Make sure the new item is visible
//
   ScrollToItem(*hlItem);

//
//  Highlight the new item
//
   DrawHighlight(hlItem, hlColor);

} // End HighlightItem

/*-----------------------------------------------------------------------
 *  Handle a pointer motion event in the area
 */

void
TreeViewC::HandlePointerMotion(Widget, TreeViewC *This, XMotionEvent*, Boolean*)
{
   if ( !This->shown || !This->focusHere ) return;

//
// Get all pending motion events
//
   XMotionEvent	mev;
   while (XCheckMaskEvent(halApp->display, PointerMotionMask, (XEvent *)&mev));

   Window	root, child;
   int		rx, ry;         // Position in root window
   int		vx, vy;         // Position in view window
   unsigned	mods;           // Keyboard modifiers

//
// Get current pointer location
//
   XQueryPointer(halApp->display, This->viewBox->ViewWin(), &root, &child,
		 &rx, &ry, &vx, &vy, &mods);

//
// Find the item under the pointer
//
   VItemC	*item = This->PickItem(vx + This->viewBox->HScrollValue(),
   				       vy + This->viewBox->VScrollValue());

   if ( item == This->hlItem ) return;

//
// Remove current highlight
//
   if ( This->hlItem ) {
      This->DrawHighlight(This->hlItem, This->bgColor);
   }
   
   This->hlItem = item;

//
//  Highlight the new item
//
   if ( This->hlItem ) {
      This->DrawHighlight(This->hlItem, This->hlColor);
   }

} // End HandlePointerMotion

/*-----------------------------------------------------------------------
 *  Change the label for an item
 */

void
TreeViewC::ChangeItemLabel(const VItemC& item)
{
//
// Look up the icon data
//
   TreeDataC	*data = (TreeDataC *)item.ViewData();
   if ( !DataValid(data) ) return;

   data->bounds.wd = 0;	// So size will be requested again

   if ( !data->Visible() ) return;

   needPlaces = True;

//
// Redraw if necessary
//
   if ( !viewBox->Deferred() ) Redraw();
   else			       viewBox->Changed(True);

} // End ChangeItemLabel

/*-----------------------------------------------------------------------
 *  Change the pixmaps for an item
 */

void
TreeViewC::ChangeItemPixmaps(const VItemC& item)
{
   //..if ( !viewBox->Realized() ) return;

//
// Look up the icon data
//
   TreeDataC	*data = (TreeDataC *)item.ViewData();
   if ( !DataValid(data) ) return;

   data->bounds.wd = 0;	// So size will be requested again

   if ( !data->Visible() ) return;

   needPlaces = True;

//
// Redraw if necessary
//
   if ( !viewBox->Deferred() ) Redraw();
   else			       viewBox->Changed(True);

} // End ChangeItemPixmaps

/*-----------------------------------------------------------------------
 *  Public method to remove an item from this view
 */

void
TreeViewC::RemoveItem(VItemC& item)
{
//
// Get icon data structure
//
   TreeDataC	*data = (TreeDataC *)item.ViewData();
   if ( !DataValid(data) ) return;

   dataDict.remove(&item);
   item.SetViewData(NULL);

   //cout <<"Deleting data " <<hex <<(int)data <<" for item "
   //	<<(int)&item SP (int)data->item <<dec NL;

   needPlaces = True;

//
// Delete icon data structure
//
   delete data;
   if ( hlItem == &item ) hlItem = NULL;

//
// Redraw if necessary
//
   if ( !viewBox->Deferred() ) Redraw();
   else			       viewBox->Changed(True);

} // End RemoveItem

/*-----------------------------------------------------------------------
 *  Remove items from this view
 */

void
TreeViewC::RemoveItems(const VItemListC& list)
{
   //..Boolean	sizeChange = False;

//
// Remove items
//
   unsigned	count = list.size();
   for (int i=0; i<count; i++) {

//
// Get icon data structure
//
      VItemC	*item = list[i];
      TreeDataC	*data = (TreeDataC *)item->ViewData();
      if ( !DataValid(data) ) continue;

      dataDict.remove(item);
      item->SetViewData(NULL);

      //cout <<"Deleting data " <<hex <<(int)data <<" for item "
      //     <<(int)item SP (int)data->item <<dec NL;

      if ( viewBox->VisItems().includes(item) ) needPlaces = True;

//
// Delete icon data structure
//
      delete data;
      if ( hlItem == item ) hlItem = NULL;

   } // End for each item to be deleted

//
// Redraw if necessary
//
   if ( !viewBox->Deferred() ) Redraw();
   else			       viewBox->Changed(True);

} // End RemoveItems

/*-----------------------------------------------------------------------
 *  Handle drag over event
 */

void
TreeViewC::HandleDragOver(XmDragProcCallbackStruct *dp)
{
   VItemC	*dropSite = PickItem(dp->x + viewBox->HScrollValue(),
   				     dp->y + viewBox->VScrollValue());
   Boolean	validDrop = (dropSite && dropSite->ValidDropSite());

   if ( dropSite != dropItem ) {

      if ( dropItem ) {
	 if ( dropItem == hlItem ) DrawHighlight(dropItem, hlColor);
	 else			   DrawHighlight(dropItem, bgColor);
      }

      if ( validDrop ) {
	 DrawHighlight(dropSite, dropColor);
	 dropItem = dropSite;
      } else {
	 dropItem = NULL;
      }

   } // End if this is a new drop site

   if ( validDrop ) dp->dropSiteStatus = XmVALID_DROP_SITE;
   else		    dp->dropSiteStatus = XmINVALID_DROP_SITE;

} // End HandleDragOver

/*-----------------------------------------------------------------------
 *  Handle drop event
 */

void
TreeViewC::DropFinished()
{
//
// Remove drop highlight
//
   if ( dropItem == hlItem ) DrawHighlight(dropItem, hlColor);
   else			     DrawHighlight(dropItem, bgColor);
   dropItem = NULL;

   if ( dragIcon ) {
      XtDestroyWidget(dragIcon);
      dragIcon = NULL;
   }

} // End DropFinished

/*-----------------------------------------------------------------------
 *  Create an icon for the drag
 */

Widget
TreeViewC::CreateDragIcon(VItemListC &list)
{
   TreeDataC	*data = (TreeDataC *)list[0]->ViewData();
   if ( !DataValid(data) ) return NULL;

//
// Create widget for drag icon
//
   WArgList	args;
#if 0
   if ( data->pixmap ) {
      args.pixmap(data->pixmap->reg);
      args.Width(data->pixmap->wd);
      args.Height(data->pixmap->ht);
   }
#endif
   dragIcon = XmCreateDragIcon(viewBox->ViewDA(), "dragIcon", ARGS);

   return dragIcon;

} // End CreateDragIcon

/*-----------------------------------------------------------------------
 *  Create an icon for the drag
 */

VItemC*
TreeViewC::DropItem()
{
   return dropItem;
}

/*-----------------------------------------------------------------------
 *  Set values
 */

void
TreeViewC::SetTreeStyle(TreeStyleT val)
{
   if ( treeStyle == val ) return;

   treeStyle  = val;
   needPlaces = True;

   if ( !viewBox->Deferred() ) Redraw();
   else			       viewBox->Changed(True);
}

void
TreeViewC::SetRootGravity(int val)
{
   if ( rootGravity == val ) return;

   rootGravity = val;
   needPlaces  = True;

   if ( !viewBox->Deferred() ) Redraw();
   else			       viewBox->Changed(True);
}

void
TreeViewC::SetGenerationSpacing(int val)
{
   if ( genSpacing == val ) return;

   genSpacing = val;
   needPlaces = True;

   if ( !viewBox->Deferred() ) Redraw();
   else			       viewBox->Changed(True);
}

void
TreeViewC::SetSiblingSpacing(int val)
{
   if ( sibSpacing == val ) return;

   sibSpacing = val;
   needPlaces = True;

   if ( !viewBox->Deferred() ) Redraw();
   else			       viewBox->Changed(True);
}

void
TreeViewC::SetLineStyle(int val)
{
   if ( lineStyle == val ) return;

   lineStyle = val;

   if ( !viewBox->Deferred() ) Redraw();
   else			       viewBox->Changed(True);
}

void
TreeViewC::SetLineWidth(int val)
{
   if ( lineWidth == val ) return;

   lineWidth = val;

   if ( !viewBox->Deferred() ) Redraw();
   else			       viewBox->Changed(True);
}

void
TreeViewC::SetCompress(Boolean val)
{
   if ( compress == val ) return;

   compress   = val;
   needPlaces = True;

   if ( !viewBox->Deferred() ) Redraw();
   else			       viewBox->Changed(True);
}

/*-----------------------------------------------------------------------
 *  Hide the specified item
 */

void
TreeViewC::Hide(VItemC *item)
{
   TreeDataC	*data = (TreeDataC *)item->ViewData();
   if ( !DataValid(data) || !data->visible ) return;

   Boolean	wasVisible = data->Visible();
   data->visible = False;

   if ( wasVisible ) {
      needPlaces = True;

      if ( !viewBox->Deferred() ) Redraw();
      else		          viewBox->Changed(True);
   }
}

/*-----------------------------------------------------------------------
 *  Hide all children of the specified item
 */

void
TreeViewC::HideChildren(VItemC *item)
{
   TreeDataC	*data = (TreeDataC *)item->ViewData();
   if ( !DataValid(data) || !item->child ) return;

   viewBox->Defer(True);

   VItemC	*child = item->child;
   while ( child ) {
      Hide(child);
      child = child->next;
   }

   viewBox->Defer(False);
}

/*-----------------------------------------------------------------------
 *  Show the specified item
 */

void
TreeViewC::Show(VItemC *item)
{
   TreeDataC	*data = (TreeDataC *)item->ViewData();
   if ( !DataValid(data) || data->visible ) return;

   data->visible = True;

   if ( data->Visible() ) {
      needPlaces = True;

      if ( !viewBox->Deferred() ) Redraw();
      else			  viewBox->Changed(True);
   }
}

/*-----------------------------------------------------------------------
 *  Show all children of the specified item
 */

void
TreeViewC::ShowChildren(VItemC *item)
{
   TreeDataC	*data = (TreeDataC *)item->ViewData();
   if ( !DataValid(data) || !item->child ) return;

   viewBox->Defer(True);

   VItemC	*child = item->child;
   while ( child ) {
      Show(child);
      child = child->next;
   }

   viewBox->Defer(False);
}

/*-----------------------------------------------------------------------
 *  Show all descendants of the specified item
 */

void
TreeViewC::ShowDescendants(VItemC *item)
{
   TreeDataC	*data = (TreeDataC *)item->ViewData();
   if ( !DataValid(data) || !item->child ) return;

   viewBox->Defer(True);

   VItemC	*child = item->child;
   while ( child ) {
      Show(child);
      ShowDescendants(child);
      child = child->next;
   }

   viewBox->Defer(False);
}

/*-----------------------------------------------------------------------
 *  Return the number of visible children below an item
 */

int
TreeViewC::VisChildCount(VItemC *item)
{
   int		count  = 0;
   VItemC	*child = item->child;
   TreeDataC	*data;
   while ( child ) {

      data = (TreeDataC *)child->ViewData();

      if ( DataVisible(data) ) count++;

      child = child->next;
   }

   return count;
}

VItemC*
TreeViewC::RootOf(VItemC *item)
{
   while ( item->parent ) item = item->parent;
   return item;
}

VItemC*
TreeViewC::FirstVisibleChild(VItemC *item)
{
   VItemC	*child = item->child;
   while ( child ) {
      TreeDataC	*data = (TreeDataC *)child->ViewData();
      if ( DataVisible(data) ) return child;
      child = child->next;
   }

   return NULL;
}

VItemC*
TreeViewC::LastVisibleChild(VItemC *item)
{
   VItemC	*last  = NULL;
   VItemC	*child = item->child;
   while ( child ) {
      TreeDataC	*data = (TreeDataC *)child->ViewData();
      if ( DataVisible(data) ) last = child;
      child = child->next;
   }

   return last;
}

