/*
 *  $Id: TreeViewModC.C,v 1.1 1998/06/24 01:29:07 lang Exp $
 *  
 *  Copyright (c) 1994 HAL Computer Systems International, Ltd.
 * 
 *          HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.
 *                  1315 Dell Avenue
 *                  Campbell, CA  95008
 *
 * Author: Greg Hilton
 * Contributors: Tom Lang, Frank Bieser, and others
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * http://www.gnu.org/copyleft/gpl.html
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#include "TreeViewModC.h"
#include "TreeViewC.h"
#include "NumericFieldC.h"
#include "WArgList.h"
#include "RowColC.h"
#include "rsrc.h"

#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/Label.h>
#include <Xm/TextF.h>
#include <Xm/RowColumn.h>
#include <Xm/ToggleB.h>

/*----------------------------------------------------------------------
 * Constructor
 */

TreeViewModC::TreeViewModC(Widget parent, char *name, ArgList argv,
			   Cardinal argc)
   : ModFormC(parent, name, argv, argc)
{
   WArgList	args;

   treeView = NULL;

//
// Create the paramForm hierarchy
//
//   paramForm
//      RowColC			paramRC
//         Label		treeStyleLabel
//         Frame		treeStyleFrame
//         Label		rootGravLabel
//         Frame		rootGravFrame
//         Label		genSpaceLabel
//         NumericFieldC	genSpaceNF
//         Label		sibSpaceLabel
//         NumericFieldC	sibSpaceNF
//         Label		lineStyleLabel
//         Frame		lineEtchFrame
//         Frame		lineDashFrame
//         Label		lineWidthLabel
//         NumericFieldC	lineWidthNF
//
   args.Reset();
   args.TopAttachment(XmATTACH_FORM);
   args.LeftAttachment(XmATTACH_FORM);
   args.RightAttachment(XmATTACH_NONE);
   args.BottomAttachment(XmATTACH_FORM);
   RowColC	*paramRC = new RowColC(paramForm, "paramRC", ARGS);

//
// Use row-major layout with 2 columns
//
   paramRC->Defer(True);
   paramRC->SetOrientation(RcROW_MAJOR);
   paramRC->SetColCount(2);
   paramRC->SetColAlignment(0, XmALIGNMENT_END);
   paramRC->SetColAlignment(1, XmALIGNMENT_BEGINNING);
   paramRC->SetColWidthAdjust(0, RcADJUST_NONE);
   paramRC->SetColWidthAdjust(1, RcADJUST_ATTACH);
   paramRC->SetColResize(0, False);
   paramRC->SetColResize(1, True);

   Widget treeStyleLabel = XmCreateLabel(*paramRC, "treeStyleLabel", 0,0);
   Widget treeStyleFrame = XmCreateFrame(*paramRC, "treeStyleFrame", 0,0);
   Widget rootGravLabel  = XmCreateLabel(*paramRC, "rootGravLabel",  0,0);
   Widget rootGravFrame  = XmCreateFrame(*paramRC, "rootGravFrame",  0,0);
   Widget genSpaceLabel  = XmCreateLabel(*paramRC, "genSpaceLabel",  0,0);
   genSpaceNF = new NumericFieldC(*paramRC, "genSpaceNF", 0,0);
   Widget sibSpaceLabel  = XmCreateLabel(*paramRC, "sibSpaceLabel",  0,0);
   sibSpaceNF = new NumericFieldC(*paramRC, "sibSpaceNF", 0,0);
   Widget lineStyleLabel = XmCreateLabel(*paramRC, "lineStyleLabel", 0,0);
   Widget lineEtchFrame  = XmCreateFrame(*paramRC, "lineEtchFrame",  0,0);
   Widget lineDashFrame  = XmCreateFrame(*paramRC, "lineDashFrame",  0,0);
   Widget lineWidthLabel = XmCreateLabel(*paramRC, "lineWidthLabel", 0,0);
   lineWidthNF = new NumericFieldC(*paramRC, "lineWidthNF", 0,0);
   compressTB = XmCreateToggleButton(*paramRC, "compressTB", 0,0);

   genSpaceNF->SetMin(0);
   sibSpaceNF->SetMin(0);
   lineWidthNF->SetMin(0);

   XtAddCallback(*genSpaceNF, XmNactivateCallback,
		 (XtCallbackProc)ChangeGenSpacing, (XtPointer)this);
   XtAddCallback(*sibSpaceNF, XmNactivateCallback,
		 (XtCallbackProc)ChangeSibSpacing, (XtPointer)this);
   XtAddCallback(*lineWidthNF, XmNactivateCallback,
		 (XtCallbackProc)ChangeLineWidth, (XtPointer)this);
   XtAddCallback(compressTB, XmNvalueChangedCallback,
		 (XtCallbackProc)ChangeCompress, (XtPointer)this);

//
// Manage widgets
//
   Widget	wlist[16];
   Cardinal	wcount = 0;
   wlist[wcount++] = treeStyleLabel;
   wlist[wcount++] = treeStyleFrame;
   wlist[wcount++] = rootGravLabel;
   wlist[wcount++] = rootGravFrame;
   wlist[wcount++] = genSpaceLabel;
   wlist[wcount++] = *genSpaceNF;
   wlist[wcount++] = sibSpaceLabel;
   wlist[wcount++] = *sibSpaceNF;
   wlist[wcount++] = lineStyleLabel;
   wlist[wcount++] = lineEtchFrame;
   wlist[wcount++] = NULL;
   wlist[wcount++] = lineDashFrame;
   wlist[wcount++] = lineWidthLabel;
   wlist[wcount++] = *lineWidthNF;
   wlist[wcount++] = NULL;
   wlist[wcount++] = compressTB;
   paramRC->AddChildren(wlist, wcount);
   paramRC->Defer(False);

//
// Create treeStyleFrame
//
//   RadioBox		treeStyleRadio
//      ToggleButton	treeAngleTB
//      ToggleButton	treeSquareTB
//      ToggleButton	treeOutlineTB
//
   args.Reset();
   args.Orientation(XmHORIZONTAL);
   args.Packing(XmPACK_TIGHT);
   Widget treeStyleRadio = XmCreateRadioBox(treeStyleFrame, "treeStyleRadio",
					    ARGS);

   treeAngleTB   = XmCreateToggleButton(treeStyleRadio, "treeAngleTB",   0,0);
   treeSquareTB  = XmCreateToggleButton(treeStyleRadio, "treeSquareTB",  0,0);
   treeOutlineTB = XmCreateToggleButton(treeStyleRadio, "treeOutlineTB", 0,0);

   wlist[0] = treeAngleTB;
   wlist[1] = treeSquareTB;
   wlist[2] = treeOutlineTB;
   XtManageChildren(wlist, 3);
   XtManageChild(treeStyleRadio);

   XtAddCallback(treeAngleTB, XmNvalueChangedCallback, 
		 (XtCallbackProc)ChangeTreeStyle, (XtPointer)this);
   XtAddCallback(treeSquareTB, XmNvalueChangedCallback, 
		 (XtCallbackProc)ChangeTreeStyle, (XtPointer)this);
   XtAddCallback(treeOutlineTB, XmNvalueChangedCallback, 
		 (XtCallbackProc)ChangeTreeStyle, (XtPointer)this);

//
// Create rootGravFrame
//
//   RadioBox		rootGravRadio
//      ToggleButton	rootNorthTB
//      ToggleButton	rootWestTB
//
   args.Reset();
   args.Orientation(XmHORIZONTAL);
   args.Packing(XmPACK_TIGHT);
   Widget rootGravRadio = XmCreateRadioBox(rootGravFrame, "rootGravRadio",
					   ARGS);

   rootNorthTB = XmCreateToggleButton(rootGravRadio, "rootNorthTB", 0,0);
   rootWestTB  = XmCreateToggleButton(rootGravRadio, "rootWestTB",  0,0);

   wlist[0] = rootNorthTB;
   wlist[1] = rootWestTB;
   XtManageChildren(wlist, 2);
   XtManageChild(rootGravRadio);

   XtAddCallback(rootNorthTB, XmNvalueChangedCallback, 
		 (XtCallbackProc)ChangeRootGrav, (XtPointer)this);
   XtAddCallback(rootWestTB, XmNvalueChangedCallback, 
		 (XtCallbackProc)ChangeRootGrav, (XtPointer)this);

//
// Create lineEtchFrame
//
//   RadioBox	lineEtchRadio
//      ToggleButton	lineFlatTB
//      ToggleButton	lineInTB
//      ToggleButton	lineOutTB
//
   args.Reset();
   args.Orientation(XmHORIZONTAL);
   args.Packing(XmPACK_TIGHT);
   Widget lineEtchRadio = XmCreateRadioBox(lineEtchFrame, "lineEtchRadio",
					   ARGS);

   lineFlatTB = XmCreateToggleButton(lineEtchRadio, "lineFlatTB", 0,0);
   lineInTB   = XmCreateToggleButton(lineEtchRadio, "lineInTB",   0,0);
   lineOutTB  = XmCreateToggleButton(lineEtchRadio, "lineOutTB",  0,0);

   wlist[0] = lineFlatTB;
   wlist[1] = lineInTB;
   wlist[2] = lineOutTB;
   XtManageChildren(wlist, 3);
   XtManageChild(lineEtchRadio);

   XtAddCallback(lineFlatTB, XmNvalueChangedCallback, 
		 (XtCallbackProc)ChangeLineStyle, (XtPointer)this);
   XtAddCallback(lineInTB, XmNvalueChangedCallback, 
		 (XtCallbackProc)ChangeLineStyle, (XtPointer)this);
   XtAddCallback(lineOutTB, XmNvalueChangedCallback, 
		 (XtCallbackProc)ChangeLineStyle, (XtPointer)this);

//
// Create lineDashFrame
//
//   RadioBox	lineDashRadio
//      ToggleButton	lineSolidTB
//      ToggleButton	lineDashTB
//
   args.Reset();
   args.Orientation(XmHORIZONTAL);
   args.Packing(XmPACK_TIGHT);
   Widget lineDashRadio = XmCreateRadioBox(lineDashFrame, "lineDashRadio",
					    ARGS);

   lineSolidTB = XmCreateToggleButton(lineDashRadio, "lineSolidTB", 0,0);
   lineDashTB  = XmCreateToggleButton(lineDashRadio, "lineDashTB",  0,0);

   wlist[0] = lineSolidTB;
   wlist[1] = lineDashTB;
   XtManageChildren(wlist, 2);
   XtManageChild(lineDashRadio);

   XtAddCallback(lineSolidTB, XmNvalueChangedCallback, 
		 (XtCallbackProc)ChangeLineStyle, (XtPointer)this);
   XtAddCallback(lineDashTB, XmNvalueChangedCallback, 
		 (XtCallbackProc)ChangeLineStyle, (XtPointer)this);

   XtManageChild(*paramRC);

} // End constructor

/*----------------------------------------------------------------------
 * Destructor
 */

TreeViewModC::~TreeViewModC()
{
   delete paramRC;
   delete genSpaceNF;
   delete sibSpaceNF;
   delete lineWidthNF;
}

/*---------------------------------------------------------------
 *  Method to apply tree changes
 */
   
void
TreeViewModC::Apply(TreeViewC& tv)
{
//
// Set the tree style
//
   if ( XmToggleButtonGetState(treeAngleTB) )
      tv.SetTreeStyle(TREE_ANGLED);
   else if ( XmToggleButtonGetState(treeSquareTB) )
      tv.SetTreeStyle(TREE_SQUARED);
   else
      tv.SetTreeStyle(TREE_OUTLINE);

//
// Set the root gravity
//
   if ( XmToggleButtonGetState(rootNorthTB) )
      tv.SetRootGravity(NorthGravity);
   else
      tv.SetRootGravity(WestGravity);

//
// Set the spacing
//
   tv.SetGenerationSpacing(*genSpaceNF);
   tv.SetSiblingSpacing   (*sibSpaceNF);

//
// Set the line style
//
   if ( XmToggleButtonGetState(lineDashTB) ) {
      if ( XmToggleButtonGetState(lineInTB) )
	 tv.SetLineStyle(XmSHADOW_ETCHED_IN_DASH);
      else if ( XmToggleButtonGetState(lineOutTB) )
	 tv.SetLineStyle(XmSHADOW_ETCHED_OUT_DASH);
      else
	 tv.SetLineStyle(XmSINGLE_DASHED_LINE);
   }

   else if ( XmToggleButtonGetState(lineInTB) )
      tv.SetLineStyle(XmSHADOW_ETCHED_IN);
   else if ( XmToggleButtonGetState(lineOutTB) )
      tv.SetLineStyle(XmSHADOW_ETCHED_OUT);
   else
      tv.SetLineStyle(XmSINGLE_LINE);

//
// Set the line width
//
   tv.SetLineWidth(*lineWidthNF);

//
// Set the compression
//
   tv.SetCompress(XmToggleButtonGetState(compressTB));

} // End Apply

/*---------------------------------------------------------------
 *  Method to initialize the settings based on the given tree
 */

void
TreeViewModC::Init(TreeViewC& tv)
{
   treeView = &tv;

//
// Save initial values
//
   init.treeStyle   = tv.TreeStyle();
   init.rootGravity = tv.RootGravity();
   init.genSpacing  = tv.GenerationSpacing();
   init.sibSpacing  = tv.SiblingSpacing();
   init.lineStyle   = tv.LineStyle();
   init.lineWidth   = tv.LineWidth();
   init.compress    = tv.Compressed();

   Reset();	// Use this to display initial values

} // End Init

/*---------------------------------------------------------------
 * Method to reset the fields to their initial values
 */

void
TreeViewModC::Reset()
{
//
// Display the initial values
//
   Widget	tb = treeOutlineTB;
   switch ( init.treeStyle ) {
      case TREE_ANGLED:		tb = treeAngleTB;	break;
      case TREE_SQUARED:	tb = treeSquareTB;	break;
      default:			tb = treeOutlineTB;	break;
   }
   XmToggleButtonSetState(tb, True, True);

   tb = (init.rootGravity == NorthGravity) ? rootNorthTB : rootWestTB;
   XmToggleButtonSetState(tb, True, True);

   Widget	tb1 = lineFlatTB;
   Widget	tb2 = lineSolidTB;
   switch ( init.lineStyle ) {
      case XmSHADOW_ETCHED_IN:	     tb1 = lineInTB;			break;
      case XmSHADOW_ETCHED_IN_DASH:  tb1 = lineInTB;  tb2 = lineDashTB; break;
      case XmSHADOW_ETCHED_OUT:	     tb1 = lineOutTB;			break;
      case XmSHADOW_ETCHED_OUT_DASH: tb1 = lineOutTB; tb2 = lineDashTB; break;
      case XmSINGLE_DASHED_LINE:		      tb2 = lineDashTB;	break;
   }
   XmToggleButtonSetState(tb1, True, True);
   XmToggleButtonSetState(tb2, True, True);

   *genSpaceNF  = init.genSpacing;
   *sibSpaceNF  = init.sibSpacing;
   *lineWidthNF = init.lineWidth;

   XmToggleButtonSetState(compressTB, init.compress, True);

   if ( autoApply ) Apply(*treeView);

} // End Reset

/*---------------------------------------------------------------
 * Callbacks to handle auto-apply
 */

void
TreeViewModC::ChangeTreeStyle(Widget tb, TreeViewModC* This,
			      XmToggleButtonCallbackStruct *data)
{
   if ( !This->autoApply || !data->set ) return;

   if ( tb == This->treeAngleTB )
      This->treeView->SetTreeStyle(TREE_ANGLED);
   else if ( tb == This->treeSquareTB )
      This->treeView->SetTreeStyle(TREE_SQUARED);
   else
      This->treeView->SetTreeStyle(TREE_OUTLINE);
}

void
TreeViewModC::ChangeLineStyle(Widget, TreeViewModC* This,
			      XmToggleButtonCallbackStruct *data)
{
   if ( !This->autoApply || !data->set ) return;

   if ( XmToggleButtonGetState(This->lineDashTB) ) {
      if ( XmToggleButtonGetState(This->lineInTB) )
	 This->treeView->SetLineStyle(XmSHADOW_ETCHED_IN_DASH);
      else if ( XmToggleButtonGetState(This->lineOutTB) )
	 This->treeView->SetLineStyle(XmSHADOW_ETCHED_OUT_DASH);
      else
	 This->treeView->SetLineStyle(XmSINGLE_DASHED_LINE);
   }

   else if ( XmToggleButtonGetState(This->lineInTB) )
      This->treeView->SetLineStyle(XmSHADOW_ETCHED_IN);
   else if ( XmToggleButtonGetState(This->lineOutTB) )
      This->treeView->SetLineStyle(XmSHADOW_ETCHED_OUT);
   else
      This->treeView->SetLineStyle(XmSINGLE_LINE);
}

void
TreeViewModC::ChangeRootGrav(Widget tb, TreeViewModC* This,
			     XmToggleButtonCallbackStruct *data)
{
   if ( !This->autoApply || !data->set ) return;

   if ( tb == This->rootNorthTB )
      This->treeView->SetRootGravity(NorthGravity);
   else
      This->treeView->SetRootGravity(WestGravity);
}

void
TreeViewModC::ChangeGenSpacing(Widget, TreeViewModC* This, XtPointer)
{
   if ( !This->autoApply ) return;

   This->treeView->SetGenerationSpacing(*This->genSpaceNF);
}

void
TreeViewModC::ChangeSibSpacing(Widget, TreeViewModC* This, XtPointer)
{
   if ( !This->autoApply ) return;

   This->treeView->SetSiblingSpacing(*This->sibSpaceNF);
}

void
TreeViewModC::ChangeLineWidth(Widget, TreeViewModC* This, XtPointer)
{
   if ( !This->autoApply ) return;

   This->treeView->SetLineWidth(*This->lineWidthNF);
}

void
TreeViewModC::ChangeCompress(Widget, TreeViewModC* This,
			     XmToggleButtonCallbackStruct *data)
{
   if ( !This->autoApply ) return;

   This->treeView->SetCompress(data->set);
}

/*---------------------------------------------------------------
 *  Method to set auto-update mode
 */

void
TreeViewModC::EnableAutoApply()
{
   autoApply = True;
}

void
TreeViewModC::DisableAutoApply()
{
   autoApply = False;
}

